/* Copyright 2010 Greg Alexander.  See the file COPYING for terms */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <readline/readline.h>
#include <readline/history.h>
#include "misc.h"
#include "lang.h"
#include "snd.h"
#include "midi.h"
#include "tracker.h"

#define MAX_MAPS 128

static void cmd_line(char *line);
static void cmd_load(char *args);
static void cmd_source(char *args);
static void cmd_map(char *args);
static void cmd_split_note(int note);
static void cmd_split(char *args);
static void cmd_show(char *args);
static void cmd_stop(char *args);
static void cmd_freq(char *args);
static void cmd_play(char *args, int loop);
static void cmd_rate(char *args);
static void cmd_dev(char *args);
static void cmd_write(char *args);
static void cmd_mic(char *args);
static void cmd_inst(char *args);
static void cmd_path(char *args);
static void cmd_cachedir(char *args);
static void cmd_sound(char *args);
static void cmd_debug(char *args);
static void cmd_gyro(char *args);
int debug_flags = 0;		/* 1 = optimizer */


void
cmd_init(void)
{
	char *home = getenv("HOME");

	search_path[0] = safe_strdup(".");
	search_path[1] = safe_strdup("..");
	if (home) 
		search_path[2] = safe_strdup(home);
	
	using_history();
	stifle_history(HIST_MAX);
	read_history(HIST_FILE);

	cmd_source(".ggbrc");

	rl_readline_name = "ggb";
/*      rl_attempted_complete_function = cmd_complete();  - XXX*/

	rl_callback_handler_install("(ggb) ", cmd_line);
}

void
cmd_do(void)
{
	rl_callback_read_char();
}

void
cmd_cleanup(void)
{
	write_history(HIST_FILE);
	rl_callback_handler_remove();
}


static char *my_getchar_buf = NULL;
static int
my_getchar(void)
{
	if (my_getchar_buf && *my_getchar_buf) {
		return *(my_getchar_buf++);
	} else {
		return EOF;
	}
}

static void
cmd_line(char *line)
{
	if (line == NULL) {
		/* EOF */
		do_cleanup();
		exit(0);
	} else {
		/* expand it */
		char *oline = line;
		int i;

		i = history_expand(line, &line);
		if (line != oline)
			free(oline);
		if (i > 0) {    /* expansion was performed, so echo it */
			printf("%s\n",line);
		}
		if (i != 0 && i != 1) {
			free(line);
			return;
		}
	}

	if (!*line) {
		free(line);
		return;
	}

	add_history(line);
	
	cmd_doit(line);
	free(line);
}

static struct tree_rec *
make_tree(char *line)
{
	struct tree_rec *tr;

	lex_getchar = my_getchar;
	my_getchar_buf = line;
	lexer_startup();
	num_errors = 0;
	tr = scan_statement();
	my_getchar_buf = NULL;
	return tr;
}

void
cmd_doit(char *line)
{
	int cmdlen = 0;

	while (isspace(*line)) line++;

	while (line[cmdlen] && !isspace(line[cmdlen])) cmdlen++;

	if (!*line || *line == '#') {
		return;
	} else if (*line == ':') {
		do_tree(make_tree(line+1));
	} else if (cmdlen && !strncmp(line, "quit", cmdlen)) {
		do_cleanup();
		exit(0);
	} else if (cmdlen && !strncmp(line, "stop", cmdlen)) {
		cmd_stop(line+cmdlen);
	} else if (cmdlen && !strncmp(line, "load", cmdlen)) {
		cmd_load(line+cmdlen);
	} else if (cmdlen && !strncmp(line, "source", cmdlen)) {
		cmd_source(line+cmdlen);
	} else if (cmdlen && !strncmp(line, "map", cmdlen)) {
		cmd_map(line+cmdlen);
	} else if (cmdlen && !strncmp(line, "split", cmdlen)) {
		cmd_split(line+cmdlen);
	} else if (cmdlen && !strncmp(line, "show", cmdlen)) {
		cmd_show(line+cmdlen);
	} else if (cmdlen && !strncmp(line, "play", cmdlen)) {
		cmd_play(line+cmdlen, /*loop=*/0);
	} else if (cmdlen && !strncmp(line, "loop", cmdlen)) {
		cmd_play(line+cmdlen, /*loop=*/1);
	} else if (cmdlen && !strncmp(line, "freq", cmdlen)) {
		cmd_freq(line+cmdlen);
	} else if (cmdlen && !strncmp(line, "rate", cmdlen)) {
		cmd_rate(line+cmdlen);
	} else if (cmdlen && !strncmp(line, "compile", cmdlen)) {
		precompile();
	} else if (cmdlen && !strncmp(line, "dev", cmdlen)) {
		cmd_dev(line+cmdlen);
	} else if (cmdlen && !strncmp(line, "help", cmdlen)) {
		help(line+cmdlen);
	} else if (cmdlen && !strncmp(line, "mrecord", cmdlen)) {
		midi_record_cmd(line+cmdlen);
	} else if (cmdlen && !strncmp(line, "mplay", cmdlen)) {
		midi_play_cmd(line+cmdlen);
	} else if (cmdlen && !strncmp(line, "write", cmdlen)) {
		cmd_write(line+cmdlen);
	} else if (cmdlen && !strncmp(line, "mic", cmdlen)) {
		cmd_mic(line+cmdlen);
	} else if (cmdlen && !strncmp(line, "inst", cmdlen)) {
		cmd_inst(line+cmdlen);
	} else if (cmdlen && !strncmp(line, "keyboard", cmdlen)) {
		keyboard_toggle();
	} else if (cmdlen && !strncmp(line, "pattern", cmdlen)) {
		pattern_cmd(line+cmdlen);
	} else if (cmdlen && !strncmp(line, "path", cmdlen)) {
		cmd_path(line+cmdlen);
	} else if (cmdlen && !strncmp(line, "cachedir", cmdlen)) {
		cmd_cachedir(line+cmdlen);
	} else if (cmdlen && !strncmp(line, "sound", cmdlen)) {
		cmd_sound(line+cmdlen);
	} else if (cmdlen && !strncmp(line, "debug", cmdlen)) {
		cmd_debug(line+cmdlen);
	} else if (cmdlen && !strncmp(line, "gyro", cmdlen)) {
		cmd_gyro(line+cmdlen);
	} else if (*line) {
		do_tree(make_tree(line));
	}
}


static char *sysex_cmds[MAX_MAPS];
static char *note_map[MAX_MAPS];

void
cmd_sysex(int number)
{
	number %= MAX_MAPS;
	if (sysex_cmds[number])
		cmd_doit(sysex_cmds[number]);
}

void
cmd_note(unsigned char note,unsigned char attack)
{
	static int active[256];
	static int bad_note = 0;

	if (note >= 128) return;

#if 0
	printf("note: %d (%d)\n",note,attack);
#endif

	if (bad_note) {
		bad_note = 0;
		cmd_note(100,0);
	}

	keyboard_note(note, attack ? 1 : 0);
	pattern_note(note, attack);
	cmd_split_note(note);
	if (gyro_note(note, attack)) {
		/* gyro reports the user failed the note, so turn it into a
		 * painful embarrassment */
		bad_note = 1;
		attack = 60;
		note = 100;
	}

		/* note up any old note */
	if (active[note]) {
		snd_noteup(active[note]);
		active[note] = 0;
	}
	if (attack != 0) {
		char buf[128];
		struct sound_rec *old = sounds;
		char *func = "inst";
		if (note_map[note])
			func = note_map[note];
		sprintf(buf,"%.80s('%s',%.3f);",func,midi_notetab[note],((double)attack)/127.0);
		cmd_doit(buf);
		if (sounds && sounds != old) {
			active[note] = sounds->id;
		}
	}
}


static FILE *my_f_getchar_f;
static int
my_f_getchar(void)
{
	if (!my_f_getchar_f)
		return EOF;
	return fgetc(my_f_getchar_f);
}

/* loads a file directly into the language parser, so multi-line
 * statements etc are valid */
static void
cmd_load(char *args)
{
	FILE *f;
	struct tree_rec *tr;

	while (isspace(*args)) args++;
	args[strcspn(args," \t\r\n")] = 0;

	if (!(f = open_search(args, "r"))) {
		perror(args);
		return;
	}

	lex_getchar = my_f_getchar;
	my_f_getchar_f = f;
	lexer_startup();
	do {
		num_errors = 0;
		tr = scan_statement();
		do_tree(tr);
	} while (curr_token != tok_eof);

	fclose(f);
	my_f_getchar_f = NULL;
}

/* loads a file, running each line as a command */
static void
cmd_source(char *args)
{
	FILE *f;
	char buf[10240];
	int n;

	while (isspace(*args)) args++;
	n = strcspn(args," \t\r\n");
	if (args[n]) args[n] = 0;

	if (!(f = open_search(args, "r"))) {
		perror(args);
		return;
	}

	while (fgets(buf,10240,f)) {
		buf[strcspn(buf,"\r\n")] = 0;
		cmd_doit(buf);
	}

	fclose(f);
}

static void
cmd_map(char *args)
{
	char **array = sysex_cmds;
	int mapno = 0;

	while (isspace(*args)) args++;

	if (!*args) {
		help("map");
		return;
	}

	if (!strncmp(args, "note", 4)) {
		/* this must be a note mapping */
		array = note_map;
		args += 4;
		while (isspace(*args)) { args++; }
	}

	if (!strncmp(args, "reset", 5)) {
		int i;
		for (i = 0; i < MAX_MAPS; i++) {
			free(array[i]);
			array[i] = NULL;
		}
		return;
	}

	/* get the note number */
	if (!isdigit(*args)) {
		help("map");
		return;
	}
	while (isdigit(*args)) { mapno = mapno * 10 + (*args-'0'); args++; }
	while (isspace(*args)) args++;
	mapno %= MAX_MAPS;

	/* remove the old one (if necessary) */
	free(array[mapno]);
	array[mapno] = NULL;

	/* kill trailing whitespace */
	if (*args) {
		char *end = args + strlen(args)-1;
		while (isspace(*end)) { *end = 0; end--; }
	}

	/* finally, save it */
	if (*args) {
		array[mapno] = safe_strdup(args);
	}
}


static int cmd_split_active = 0;	/* 0 = inactive, 1 = up, 2 = down */
static char *cmd_split_inst = NULL;

static void
cmd_split_note(int note)
{
	int i = 0, delta = 0;

	if (cmd_split_active == 1) {
		delta = 1;
	} else if (cmd_split_active == 2) {
		delta = -1;
	} else {
		return;
	}

	for (i = note; i < MAX_MAPS && i >= 0; i += delta) {
		free(note_map[i]);
		note_map[i] = NULL;

		if (cmd_split_inst) {
			note_map[i] = safe_strdup(cmd_split_inst);
		}
	}

	free(cmd_split_inst);

	cmd_split_active = 0;
	cmd_split_inst = NULL;
}

static void
cmd_split(char *args)
{
	int direction = 0;

	while (isspace(*args)) args++;

	if (!strncmp(args, "up", 2)) {
		direction = 1;
		args += 2;
	} else if (!strncmp(args, "down", 4)) {
		direction = 2;
		args += 4;
	} else {
		help("split");
	}

	while (isspace(*args)) args++;

	if (!*args) {
		cmd_split_inst = NULL;
	} else {
		cmd_split_inst = safe_strdup(args);
	}

	cmd_split_active = direction;

	printf("play a note, then all notes %s than it will be remapped.\n", ((direction == 1) ? "higher" : "lower"));
}

static void
cmd_show(char *args)
{
	int cmdlen = 0;
	while (isspace(*args)) args++;
	cmdlen = strcspn(args," \t\r\n");
	if (cmdlen && !strncmp(args,"sounds",cmdlen)) {
		struct sound_rec *s;
		printf("snd_max = %.3f\n\n", ((double)snd_max) / 32767.0);
		for (s = sounds; s; s = s->next) {
			printf("sound at ofs=%d, note_up=%d:\n",(int)s->ofs,(int)s->note_up);
			printf("   ");
			dump_tree(s->tr);
			printf("\n\n");
		}
	} else if (cmdlen && !strncmp(args,"functions",cmdlen)) {
		struct symbol *s;
		for (s = symtab; s; s = s->next) {
			if (!s->defn)
				continue;
			assert_str(s->defn->kind == op_fn_def, "fn defined with non-fn_def");
			printf("func %s%s:\n", s->name, s->compiled ? " (linked)" : "");
			printf("   ");
			dump_tree(s->defn);
			printf("\n\n");
		}
	} else if (cmdlen && !strncmp(args,"maps",cmdlen)) {
		int i;
		char *last = NULL;
		int last_dots = 0;

		printf("sysex command maps:\n");
		for (i = 0; i < MAX_MAPS; i++) {
			if (sysex_cmds[i])
				printf("   sysex %d: %s\n",i,sysex_cmds[i]);
		}
		printf("note function maps:\n");
		for (i = 0; i < MAX_MAPS; i++) {
			if (!note_map[i]) {
				if (last) {
					printf("   note %d: %s\n",i-1,last);
				}
				last_dots = 0;
			} else if (last && !strcmp(note_map[i], last)) {
				if (!last_dots) {
					printf("   ...\n");
					last_dots = 1;
				}
			} else {
				if (last && last_dots) {
					printf("   note %d: %s\n", i-1, last);
				}
				printf("   note %d: %s\n",i,note_map[i]);
				last_dots = 0;
			}
			last = note_map[i];
		}
		if (last && last_dots) {
			printf("   note %d: %s\n",i-1,last);
		}
	} else if (cmdlen && !strncmp(args,"tracker",cmdlen)) {
		tracker_show_status();
	} else if (cmdlen && !strncmp(args,"midi",cmdlen)) {
		midi_status_cmd();
	} else {
		help("show");
	}
}

static void
cmd_stop(char *args)
{
	int cmdlen;
	while (isspace(*args)) args++;
	cmdlen = strcspn(args," \t\r\n");
	if (cmdlen && !strncmp(args,"sounds",cmdlen)) {
		snd_shutup();
	} else if (cmdlen && !strncmp(args,"trackers",cmdlen)) {
		tracker_stop();
	} else if (cmdlen && !strncmp(args,"midi", cmdlen)) {
		midi_stop_cmd();
	} else if (cmdlen && !strncmp(args,"pattern", cmdlen)) {
		pattern_stop();
	} else if (cmdlen && !strncmp(args,"gyro", cmdlen)) {
		gyro_stop();
	} else if (cmdlen && !strncmp(args,"all",cmdlen)) {
		snd_shutup();
		tracker_stop();
		midi_stop_cmd();
		pattern_stop();
		gyro_stop();
	} else {
		help("stop");
	}
}

static void
cmd_freq(char *args)
{
	int steps = 0;
	double base = 0.0;
	int i;
	double diff = 0.0;
	double total = 0.0;

	while (isspace(*args)) args++;
	steps = strtoul(args, &args, 10);
	while (isspace(*args)) args++;
	base = strtod(args, NULL);
	if (!steps || !base) {
		help("freq");
		return;
	}

	for (i = 0; i < 128; i++) {
		midi_freqtab[i] = base * pow(2.0,((double)i) / ((double)steps));
		diff += (midi_freqtab[i] - midi_freqtab_base[i]);
		total += fabs(midi_freqtab[i] - midi_freqtab_base[i]);
	}
	diff /= 128.0; total /= 128.0;
	printf("difference: %.4f, total: %.4f\n",diff,total);
}

static void
cmd_play(char *args, int loop)
{
	char *fn = NULL;
	char *skip = NULL;

	while(isspace(*args)) args++;
	fn = args;
	while (*args) {
		if (isspace(*args)) {
			*args = 0;
			args++;
			skip = args;
		}
		args++;
	}

	if (!*fn) {
		help(loop ? "loop" : "play");
		return;
	}

	tracker_play(fn, loop, skip ? atoi(skip) : 0);
}

static void
cmd_rate(char *args)
{
	unsigned rate;
	
	while (isspace(*args)) args++;
	if (!*args) {
		help("rate");
		return;
	}
	rate = atoi(args);
	if (rate) tracker_rate(rate);
}

static void
cmd_dev(char *args)
{
	args += strspn(args," \t");
	if (!*args) args = NULL;
	snd_changedev(args);
}

static void
cmd_write(char *line)
{
	if (snd_file_fd != -1) {
		double max = ((double)snd_max) / 32767.0;
		printf("closing output, 1/max = %.2f\n\n",1.0/max);
		close(snd_file_fd);
		snd_file_fd = -1;
	}
	line += strspn(line," \t");
	if (*line) {
		snd_file_fd = open(line, O_WRONLY|O_CREAT, 0666);
		if (snd_file_fd == -1) {
			perror(line);
		} else {
			snd_max = 0;
			printf("outputting to %s\n",line);
		}
	}
}

static void
cmd_mic(char *line)
{
	double d;
	line += strspn(line," \t");
	d = strtod(line, NULL);
	mic_level = (d * 32767.0);
}

struct inst_list_rec {
	struct inst_list_rec *next;
	struct inst_list_rec *prev;
	char expression[0];
};
static struct inst_list_rec *inst_list = NULL, *inst_list_curr = NULL;

static void
do_inst(struct inst_list_rec *t)
{
	static char *s = NULL;
	static int s_len = 0;
	int saved_nag_warnings = nag_warnings;

	if (!t) return;

	if (strlen(t->expression) + 20 > s_len) {
		free(s);
		s_len = strlen(t->expression) + 20;
		s = safe_malloc(s_len);
	}
	sprintf(s,"inst[x,y]=%s;",t->expression);

	nag_warnings = 0;
	do_tree(make_tree(s));
	nag_warnings = saved_nag_warnings;

	printf("\rswitching to: %s    \n",t->expression);
	inst_list_curr = t;
}

static void
cmd_inst(char *line)
{
	int cmdlen = 0;
	line += strspn(line," \t");
	while (isalpha(line[cmdlen])) cmdlen++;
	if (cmdlen && !strncmp(line, "add", cmdlen)) {
		struct inst_list_rec *n;
		char *s = line+cmdlen;
		int len;
		s += strspn(s," \t");
		len = strlen(s);
		if (!len) {
			help("inst");
			return;
		}
		n = malloc(sizeof *n + len + 1);
		n->next = NULL;
		strcpy(n->expression, s);
		if (n->expression[len-1] == ';') {
			n->expression[len-1] = 0;
		}
		if (inst_list) {
			n->prev = inst_list->prev;
			n->next = inst_list;
			inst_list->prev->next = n;
			inst_list->prev = n;
		} else {
			n->next = n->prev = n;
			inst_list = n;
		}
	} else if (cmdlen && !strncmp(line, "next", cmdlen)) {
		do_inst(inst_list_curr ? inst_list_curr->next : inst_list);
	} else if (cmdlen && !strncmp(line, "prev", cmdlen)) {
		do_inst(inst_list_curr ? inst_list_curr->prev : inst_list);
	} else if (cmdlen && !strncmp(line, "show", cmdlen)) {
		struct inst_list_rec *t;
		int count = 0;
		printf("instrument list:\n");
		for (t = inst_list; t && (!count || t != inst_list); t = t->next) {
			printf(" %3d%s: %s\n", count, ((t == inst_list_curr) ? "*" : ""), t->expression);
			count++;
		}
	} else if (cmdlen && !strncmp(line, "goto", cmdlen)) {
		struct inst_list_rec *t;
		char *s = line+cmdlen;
		int count;
		if (!inst_list) return;
		s += strspn(s," \t");
		count = atoi(s);
		for (t = inst_list; t && count; t = t->next) {
			count--;
		}
		do_inst(t);
	} else if (cmdlen && !strncmp(line, "reset", cmdlen)) {
		struct inst_list_rec *t, *t_next;
		inst_list_curr = NULL;
		if (inst_list)
			inst_list->prev->next = NULL;
		for (t = inst_list; t; t = t_next) {
			t_next = t->next;
			free(t);
		}
		inst_list = NULL;
	} else {
		help("inst");
	}
}

/* returns the current instrument function name.. */
char *
cmd_inst_getname(void)
{
	static char *buf = NULL;
	static int buflen = 0;

	if (!inst_list_curr) {
		return "inst";
	}

	if (strlen(inst_list_curr->expression)+1 > buflen) {
		free(buf);
		buflen = strlen(inst_list_curr->expression)+100;
		buf = safe_malloc(buflen);
	}

	strcpy(buf, inst_list_curr->expression);
	buf[strcspn(buf,"(")] = 0;
	if (!buf[0]) {
		return "inst";
	}

	return buf;
}

static void
cmd_path(char *line)
{
	line += strspn(line," \t");
	if (!*line) {
		/* display the path */
		int i;
		for (i = 0; i < SEARCH_MAX; i++) {
			if (!search_path[i])
				continue;
			printf("%s%s",(i == 0 ? "" : ":"), search_path[i]);
		}
		printf("\n");
	} else {
		char *eop;
		char *next;
		int n = 0;

		memset(search_path, 0, sizeof search_path);

		while (*line) {
			eop = line + strcspn(line, ":\r\n");
			next = (*eop ? eop+1 : eop);
			*eop = 0;
			if (*line) {
				search_path[n++] = safe_strdup(line);
			}
			line = next;
		}
	}
}

static void
cmd_cachedir(char *line)
{
	line += strspn(line," \t");
	if (!*line) {
		/* display the cache dir */
		if (cache_dir) {
			printf("  cache directory: %s\n",cache_dir);
		} else {
			printf("  not saving .so files into cache\n");
		}
	} else {
		if (cache_dir) free(cache_dir);
		cache_dir = safe_strdup(line);
	}
}

static void
cmd_sound(char *line)
{
	line += strspn(line," \t");
	if (!strncmp(line,"off",3)) {
		if (snd_fd != -1) snd_off();
	} else if (!strncmp(line,"on",2)) {
		if (snd_fd == -1) snd_on();
	} else {
		help("sound");
	}
}


static void
cmd_debug(char *args)
{
	while (isspace(*args)) args++;
	debug_flags = atoi(args);
}

static void
cmd_gyro(char *args)
{
	int bpm = -1;
	int mode = 1;
	int ofs = 0;
	char *fn;

	args += strspn(args," ");
	fn = args;
	args += strcspn(args," ");
	if (*args) {
		*args++ = 0;
		args += strspn(args," ");
	}
	if (*args) {
		if (isdigit(*args)) {
			bpm = strtoul(args, &args, 10);
			args += strspn(args," ");
			if (*args) {
				if (isdigit(*args)) {
					mode = strtoul(args, &args, 10);
					args += strspn(args, " ");
					if (*args) {
						if (isdigit(*args)) {
							ofs = strtoul(args, &args, 10);
						} else {
							printf("invalid offset argument\n");
						}
					}
				} else {
					printf("invalid mode argument\n");
				}
			}
		} else {
			printf("invalid bpm argument\n");
			help("gyro");
			return;
		}
	}
	gyro_start(fn, bpm, mode, ofs);
}
