/* Copyright 2010 Greg Alexander.  See the file COPYING for terms */
/* $Id: lang.c,v 1.39 2010-02-08 03:17:28 greg Exp $ */
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>
#include <math.h>
#include "lang.h"
#include "misc.h"


value_t
func_graph(struct sound_rec *curr_sound, value_t *tmps, intval_t *itmps,value_t x,value_t y,value_t n1,value_t n2,value_t n3,value_t n4,value_t n5,value_t n6,int *done)
{
	static FILE *f = NULL;
	if (!f) {
		f = fopen("graphout","w");
	}
	if (f) {
		fprintf(f,"%f %f\n",x,y);
	}

	return y;
}


/* a helper function/datatype for func_waveform()...  wave_get reads a
 * wave file into the cache and returns a pointer to its cache entry */

#define WAVEFORM_HASH_SIZE 41

struct raw_waveform {
	struct raw_waveform *next;
	char *name;
	value_t *data;
	unsigned long len;
};
static struct raw_waveform *wave_cache[WAVEFORM_HASH_SIZE];

static unsigned
hash_string(char *s)
{
	unsigned ret = 0;

	while (*s) {
		ret = ret * 37 + *s;
		s++;
	}

	return ret;
}

#define WAVE_BUFSZ 1024

static struct raw_waveform *
wave_read(char *fn)
{
	FILE *f;
	struct raw_waveform *ret;
	signed short buf[WAVE_BUFSZ];
	unsigned ofs;
	int read_words;

	f = fopen(fn, "r");
	if (!f) {
		perror(fn);
		return NULL;
	}

	ret = safe_malloc(sizeof *ret);
	ret->next = NULL;
	ret->name = safe_strdup(fn);

	fseek(f, 0, SEEK_END);
	ret->len = ftell(f) / 2;
	fseek(f, 0, SEEK_SET);

	ret->data = safe_malloc(ret->len * sizeof *(ret->data));

	for (ofs = 0; ofs < ret->len; ofs += read_words) {
		int i;
		read_words = fread(buf, sizeof buf[0], WAVE_BUFSZ, f);
		if (read_words <= 0) {
			perror("fread");
			break;
		}
		for (i = 0; i < read_words; i++) {
#if __BYTE_ORDER == __BIG_ENDIAN
			/* the file format is little endian, so convert... */
			buf[i] = ((buf[i] & 0xff00) >> 8) |
			       ((buf[i] & 0x00ff) << 8);
#endif
			ret->data[ofs+i] = ((value_t)buf[i]) / 32768.0;
		}
	}

	fclose(f);

	return ret;
}

struct raw_waveform *
wave_get(char *fn)
{
	unsigned bucket = hash_string(fn) % WAVEFORM_HASH_SIZE;
	struct raw_waveform *wav;

	for (wav = wave_cache[bucket]; wav; wav = wav->next) {
		if (!strcmp(wav->name, fn))
			break;
	}

	if (wav) return wav;

	wav = wave_read(fn);

	if (!wav) return NULL;

	wav->next = wave_cache[bucket];
	wave_cache[bucket] = wav;
	return wav;
}



static void
add_builtin(char *name, int num_args, int num_temps, int num_itemps, int wave)
{
	struct symbol *sym;
	struct builtin_rec *bi;

	sym = sym_lookup(name);
	sym->defn = NULL;
	bi = (struct builtin_rec *)safe_malloc(sizeof *bi);
	bi->num_args = num_args;
	bi->num_temps = num_temps;
	bi->num_itemps = num_itemps;
	bi->wave = wave;
	sym->builtin = bi;
}

void
lang_init(void)
{
	add_builtin("sin", 1, 0, 0, WAV_FREQ);
	add_builtin("cos", 1, 0, 0, WAV_FREQ);
	add_builtin("white", 0, 0, 0, WAV_FREQ);
	add_builtin("exp", 2, 0, 0, 0);
	add_builtin("graph", 2, 0, 0, 0);
	add_builtin("waveform", 2, 2, 0, WAV_FREQ);
}

void
reset_compileds(void)
{
	struct symbol *sym;
	for (sym = symtab; sym; sym = sym->next) {
		if (sym->compiled && sym->defn) {
			compiled_free(sym->compiled);
			sym->compiled = NULL;
		}
	}
}

void
lang_cleanup(void)
{
	struct symbol *sym, *sym_next;
	for (sym = symtab; sym; sym = sym_next) {
		sym_next = sym->next;
		if (sym->compiled) {
			compiled_free(sym->compiled);
			sym->compiled = NULL;
		}
		free(sym->name);
		free_tree(sym->defn);
		free(sym);
	}
}

struct symbol *
sym_lookup(char *s)
{
	struct symbol *sym;
	for (sym = symtab; sym; sym = sym->next) {
		if (!strcmp(sym->name,s))
			return sym;
	}
	sym = safe_malloc(sizeof *sym);
	memset(sym,0,sizeof *sym);
	sym->next = symtab;
	sym->name = safe_strdup(s);
	symtab = sym;
	return sym;
}

enum t_kind
is_t_sym(struct symbol *sym)
{
	if (sym->name[0] == 't') {
		if (sym->name[1] == 0) {
			return t_default;
		}
	}
	return t_none;
}

void
do_error(char *fmt,...)
{
	va_list ap;

	num_errors++;

	va_start(ap,fmt);
	fprintf(stderr,"error: ");
	vfprintf(stderr,fmt,ap);
	va_end(ap);
	fprintf(stderr,"\n");
}

void
do_warning(char *fmt,...)
{
	va_list ap;

	num_warnings++;

	va_start(ap,fmt);
	fprintf(stderr,"warning: ");
	vfprintf(stderr,fmt,ap);
	va_end(ap);
	fprintf(stderr,"\n");
}

static char *lang_strtab = NULL;
static int lang_strtab_used = 0;
static int lang_strtab_size = 0;

int
add_string_to_strtab(char *s)
{
	int slen = strlen(s);
	int ret;

	if (slen + lang_strtab_used + 10 > lang_strtab_size) {
		char *new_strtab;
		int new_size;
		new_size = slen + lang_strtab_used + 4000;
		new_strtab = safe_malloc(new_size);
		memcpy(new_strtab, lang_strtab, lang_strtab_size);
		memset(new_strtab+lang_strtab_size, 0,
			new_size-lang_strtab_size);
		free(lang_strtab);
		lang_strtab = new_strtab;
		lang_strtab_size = new_size;
	}
	ret = lang_strtab_used;
	memcpy(lang_strtab+lang_strtab_used, s, slen+1);
	lang_strtab_used += slen + 1;
	return ret;
}

char *
get_string_from_strtab(int idx)
{
	if (idx < 0) return "";
	if (idx >= lang_strtab_size) return "";
	return lang_strtab+idx;
}
