/* Copyright 2010 Greg Alexander.  See the file COPYING for terms */
/* $Id: lang.h,v 1.50 2010-02-08 03:17:28 greg Exp $ */
#ifndef LANG_H
#define LANG_H

#if DEFINE_GLOBALS
#define __EXTERN__
#define EQUALS(x) = (x)
#else
#define __EXTERN__ extern
#define EQUALS(x)
#endif

/* the following constraints seem to be in effect:
 * FREQ_MULT = the number of points on our interpolated sinewave
 * SEC_MULT = the number of actual samples in one second of output */

#ifdef MAEMO
#define FREQ_MULT 1024.0
#else
#define FREQ_MULT 32768.0
#endif
#define SEC_MULT 44100.0
typedef double value_t;
typedef struct { value_t pan, value; } svalue_t;
typedef signed int intval_t;

struct sound_rec;

typedef value_t (*builtin_func)(struct sound_rec*,value_t*,intval_t*,value_t,value_t,value_t,value_t,value_t,value_t,value_t,value_t,int *);
typedef int (*compiled_func)(struct sound_rec*,value_t*,intval_t*,value_t,value_t,value_t,value_t,value_t,value_t,value_t,value_t,int *buf,intval_t nents);

struct builtin_rec {
	int num_args;
	int num_temps;
	int num_itemps;
	int wave;
#define WAV_FREQ	0x01	/* has period FREQ_MULT */
};
struct compiled_rec {
	compiled_func func;
	int num_args;
	int num_temps;
	int num_itemps;
	void *dlh;
};

struct symbol {
	struct symbol *next;
	char *name;
	struct tree_rec *defn;
	struct builtin_rec *builtin;
	struct compiled_rec *compiled;
	int formal_index;
	int formal_func;
} *symtab;

typedef enum {
	tok_error,
	tok_eof,
	tok_semicolon,
	tok_lparen,
	tok_rparen,
	tok_lbracket,
	tok_rbracket,
	tok_lbrace,
	tok_rbrace,
	tok_comma,
	tok_equals,
	tok_colon,
	tok_coloncolon,
	tok_plus,
	tok_minus,
	tok_star,
	tok_slash,
	tok_symbol,
	tok_number,
	tok_string
} token;


typedef enum {
	op_error,
	op_fn_def,
	op_fn_call,
	op_add,
	op_subtract,
	op_multiply,
	op_divide,
	op_termseq,
	op_number,
	op_argref,
	op_string,
	op_timest,
	op_stereo,
	op_one_tree,
	op_raised
} tree_op;

struct tree_rec {
	tree_op kind;
	union {
		/* op_error gets nothing */
		struct {		/* op_fn_def */
			struct symbol *name;
			struct symbol **formals;
			int num_formals;
			struct tree_rec *expr;
			struct tree_rec *opt_expr;	/* after opt.c */
		} fn_def;
		struct {		/* op_fn_call */
			struct symbol *func;
			struct tree_rec **args;
			int num_args;
			int tmp_index;
			int itmp_index;
		} fn_call;
				/* op_add,op_subtract,op_multiply,op_divide */
		struct tree_rec *operands[2];
		struct {		/* op_termseq */
			struct tree_rec **durations;
			struct tree_rec **exprs;
			char *zero_crossing;
			int num_intervals;
			int repeats;
		} termseq;
		value_t number;		/* op_number */
		int string;		/* op_string */
		int argref;		/* op_argref */
#define ARG_RAISED_FORMAL		0x40000000 /* reference raised formal */
#define ARG_RAISED_COMMON		0x20000000 /* reference raised common */
		struct {                /* op_stereo */
			struct tree_rec *pan,*value;
		} stereo;
		struct {                /* op_timest */
			struct tree_rec *operand;
			enum t_kind {
				t_none = 0,
				t_default,      /* user value of some sort */
				t_freq          /* argument to builtin */
			} kind;
		} timest;
		struct {		/* op_one_tree */
			struct tree_rec *expr;
			value_t *tmps;
			intval_t *itmps;
			int num_temps;
			int num_itemps;
		} one_tree;
		struct {		/* op_raised */
			struct tree_rec *expr;
			struct tree_rec **formals;	/* depends on formals */
			int num_formals;
			struct tree_rec **commons;	/* common exprs */
			int num_commons;
		} raised;
	} u;
};


__EXTERN__ token curr_token EQUALS(tok_error);
__EXTERN__ struct symbol *curr_token_sym;	/* for tok_symbol */
__EXTERN__ char *curr_token_str;		/* for tok_string */
__EXTERN__ double curr_token_num;		/* for tok_number */

__EXTERN__ int (*lex_getchar)(void);	

__EXTERN__ int num_errors EQUALS(0);
__EXTERN__ int num_warnings EQUALS(0);
__EXTERN__ int nag_warnings EQUALS(1);

/* lexer.c: */
void lexer_startup(void);
token next_token(void);
char *dump_token(void);

/* tree.c: */
struct tree_rec *new_tree(tree_op kind);
void free_tree(struct tree_rec *tr);
struct tree_rec *dup_tree(struct tree_rec *tr);
struct tree_rec *scan_statement(void);
void dump_tree(struct tree_rec *tr);
int tree_equal(struct tree_rec *x, struct tree_rec *y);

/* lang.c: */
struct raw_waveform *wave_get(char *fn);
void lang_init(void);
void reset_compileds(void);
void lang_cleanup(void);
struct symbol *sym_lookup(char *s);
enum t_kind is_t_sym(struct symbol *sym);
void do_error(char *,...);
void do_warning(char *,...);
int add_string_to_strtab(char *s);
char *get_string_from_strtab(int idx);
value_t func_graph(struct sound_rec *curr_sound, value_t *tmps, intval_t *itmps,value_t x,value_t y,value_t n1,value_t n2,value_t n3,value_t n4,value_t n5,value_t n6,int *done);

/* sinetable.c: */
__EXTERN__ value_t *sine_table EQUALS(NULL);
void sinetable_init(void);

/* eval.c: */
int do_fn_call(struct sound_rec *s, int *buf, int nents);
void do_tree(struct tree_rec *tr);

/* compile.c: */
__EXTERN__ char *cache_dir EQUALS(NULL);
struct compiled_rec *compile_func(struct tree_rec *tr);
void precompile(void);
void compiled_free(struct compiled_rec *bi);

/* opt.c: */
void opt(struct tree_rec *tr);

#undef __EXTERN__
#undef EQUALS

#endif /* LANG_H */
