/* Copyright 2010 Greg Alexander.  See the file COPYING for terms */
/* $Id: main.c,v 1.27 2010-02-08 03:17:28 greg Exp $ */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <errno.h>
#include <signal.h>

#define DEFINE_GLOBALS 1
#include "lang.h"
#include "snd.h"
#include "midi.h"
#include "misc.h"
#include "tracker.h"


static void select_loop(void);
static void report_usage(void);
static time_t tstart;

static void sighand(int sig);

int
main(int argc,char **argv)
{
	time(&tstart);
	srandom(0);		/* a canonical state for white noise */

	sinetable_init();
	lang_init();
	snd_init();
	midi_init();
	cmd_init();

		/* execute commandline arguments as though typed at prompt */
	if (argc > 1) {
		char *buf;
		int buflen = 0;
		int i;
		for (i = 1; i < argc; i++) {
			buflen += strlen(argv[i])+2;
		}
		buf = safe_malloc(buflen+4);
		*buf = 0;
		for (i = 1; i < argc; i++) {
			strcat(buf,argv[i]);
			strcat(buf," ");
		}
		cmd_doit(buf);
	}

	signal(SIGFPE, sighand);

	select_loop();

	do_cleanup();
	return 0;
}

static void
select_loop(void)
{
	fd_set rfds,wfds;
	int i;
	int max_fd = 0;
	struct timeval tv, *tvp;

	while (1) {
		tvp = NULL;
		FD_ZERO(&rfds);
		FD_ZERO(&wfds);

		FD_SET(0,&rfds);
		if (midi_fd != -1) {
			FD_SET(midi_fd,&rfds);
			if (midi_fd > max_fd) max_fd = midi_fd;
		}
		if (snd_fd != -1) {
			/* wait until sound can be output to */
			FD_SET(snd_fd,&wfds);
			if (snd_fd > max_fd) max_fd = snd_fd;
		} else if (snd_file_fd != -1) {
			/* don't wait unless somehow snd_file_fd blocks */
			FD_SET(snd_file_fd,&wfds);
			if (snd_file_fd > max_fd) max_fd = snd_file_fd;
		} else if (sounds != NULL || tracker_playing()) {
			/* don't wait at all, sounds active */
			tv.tv_usec = 0;
			tv.tv_sec = 0;
			tvp = &tv;
		} else {
			/* wait for something else to happen */
		}
		if (keyboard_x_fd != -1) {
			FD_SET(keyboard_x_fd,&rfds);
			if (keyboard_x_fd > max_fd) max_fd = keyboard_x_fd;
		}
		if (gyro_x_fd != -1) {
			FD_SET(gyro_x_fd,&rfds);
			if (gyro_x_fd > max_fd) max_fd = gyro_x_fd;
		}

		i = select(max_fd+1,&rfds,&wfds,NULL,tvp);
		if (i < 0) {
			if (errno != EINTR)
				perror("select");
			continue;
		}
		if (FD_ISSET(0,&rfds)) {
			cmd_do();
		}
		if (midi_fd != -1 && FD_ISSET(midi_fd,&rfds)) {
			midi_in();
		}
		if (snd_fd != -1 && FD_ISSET(snd_fd,&wfds)) {
			snd_out();
		} else if (snd_file_fd != -1 && FD_ISSET(snd_file_fd,&wfds)) {
			snd_out();
		} else if ((snd_fd == -1) && (snd_file_fd == -1) &&
		           (sounds != NULL || tracker_playing())) {
			snd_out();
		}
		if (keyboard_x_fd != -1 && FD_ISSET(keyboard_x_fd,&rfds)) {
			keyboard_poll();
		}
		if (gyro_x_fd != -1 && FD_ISSET(gyro_x_fd,&rfds)) {
			gyro_poll();
		}
	}
}


void
do_cleanup(void)
{
	if (gyro_x_fd != -1) {
		gyro_cleanup();
	}
	if (keyboard_x_fd != -1) {
		keyboard_toggle();
	}
	tracker_cleanup();
	cmd_cleanup();
	midi_cleanup();
	snd_cleanup();
	lang_cleanup();
	report_usage();
}

static void
report_usage(void)
{
	struct rusage rself,rchild;
	int utself,utchild,stself,stchild;
	int total;
	time_t tend;
	getrusage(RUSAGE_SELF,&rself);
	getrusage(RUSAGE_CHILDREN,&rchild);
	time(&tend);
	utself = (rself.ru_utime.tv_sec * 10) + (rself.ru_utime.tv_usec / 100000);
	utchild = (rchild.ru_utime.tv_sec * 10) + (rchild.ru_utime.tv_usec / 100000);
	stself = (rself.ru_stime.tv_sec * 10) + (rself.ru_stime.tv_usec / 100000);
	stchild = (rchild.ru_stime.tv_sec * 10) + (rchild.ru_stime.tv_usec / 100000);
	total = tend - tstart;
	printf("%.1fu+%.1fs (gcc: %.1fu+%.1fs) out of %d (%.1f%%)\n",utself/10.0,stself/10.0,utchild/10.0,stchild/10.0,total,((utself+stself+utchild+stchild)/10.0)*100.0/((double)total));
}

static void
sighand(int sig)
{
	signal(sig,sighand);
	if (sig == SIGFPE) {
		fprintf(stderr,"SIGFPE\n");
	} else {
		fprintf(stderr,"SIG%d\n",sig);
	}
}
