/* Copyright 2010 Greg Alexander.  See the file COPYING for terms */
/* $Id: midi.c,v 1.26 2010-02-08 03:17:28 greg Exp $ */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/time.h>
#if defined(OSS_MIDI)
# include <sys/soundcard.h>
#elif defined(COREAUDIO_MIDI)
# include <CoreMIDI/MIDIServices.h>
# include <CoreFoundation/CFRunLoop.h>
#endif
#include "lang.h"
#include "midi.h"
#include "snd.h"
#include "misc.h"


double midi_freqtab[128];

const double midi_freqtab_base[128] = {
8.18,8.66,9.18,9.72,10.30,10.91,11.56,12.25,12.98,
13.75,14.57,15.43,16.35,17.32,18.35,19.45,20.60,21.83,23.12,24.50,25.96,
27.50,29.14,30.87,32.70,34.65,36.71,38.89,41.20,43.65,46.25,49.00,51.91,
55.00,58.27,61.73,65.41,69.30,73.42,77.78,82.41,87.31,92.50,98.00,103.83,
110.00,116.54,123.47,130.81,138.59,146.83,155.56,164.81,174.62,185.00,196.00,207.65,
220.00,233.08,246.94,261.62,277.19,293.67,311.12,329.63,349.23,370.00,392.00,415.31,
440.00,466.16,493.88,523.25,554.37,587.33,622.25,659.26,698.46,739.99,783.99,830.61,
880.00,932.32,987.76,1046.50,1108.74,1174.66,1244.50,1318.52,1396.92,1479.98,1567.98,1661.22,
1760.00,1864.64,1975.52,2093.00,2217.48,2349.32,2489.00,2637.04,2793.84,2959.96,3135.96,3322.44,
3520.00,3729.28,3951.04,4186.00,4434.96,4698.64,4978.00,5274.08,5587.68,5919.92,6271.92,6644.88,
7040.00,7458.56,7902.08,8372.00,8869.92,9397.28,9956.00,10548.16,11175.36,11839.84,12543.84
};
const char midi_notetab[128][4] = {
"C-0","C#0","D-0","D#0","E-0","F-0","F#0","G-0","G#0","A-0","A#0","B-0", /*00*/
"C-1","C#1","D-1","D#1","E-1","F-1","F#1","G-1","G#1","A-1","A#1","B-1", /*0C*/
"C-2","C#2","D-2","D#2","E-2","F-2","F#2","G-2","G#2","A-2","A#2","B-2", /*18*/
"C-3","C#3","D-3","D#3","E-3","F-3","F#3","G-3","G#3","A-3","A#3","B-3", /*24*/
"C-4","C#4","D-4","D#4","E-4","F-4","F#4","G-4","G#4","A-4","A#4","B-4", /*30*/
"C-5","C#5","D-5","D#5","E-5","F-5","F#5","G-5","G#5","A-5","A#5","B-5", /*3Cm*/
"C-6","C#6","D-6","D#6","E-6","F-6","F#6","G-6","G#6","A-6","A#6","B-6", /*48*/
"C-7","C#7","D-7","D#7","E-7","F-7","F#7","G-7","G#7","A-7","A#7","B-7", /*54*/
"C-8","C#8","D-8","D#8","E-8","F-8","F#8","G-8","G#8","A-8","A#8","B-8", /*60*/
"C-9","C#9","D-9","D#9","E-9","F-9","F#9","G-9","G#9","A-9","A#9","B-9", /*6C*/
"C-A","C#A","D-A","D#A","E-A","F-A","F#A","G-A" };                       /*78*/


static void midi_record_event(char kind, int x, int y);


int
midi_init(void)
{
	memcpy(midi_freqtab,midi_freqtab_base,sizeof midi_freqtab);

#if defined(OSS_MIDI)
	midi_fd = open(MIDI_DEVICE,O_RDONLY);
	if (midi_fd == -1) {
		perror(MIDI_DEVICE);
		return -1;
	}
#elif defined(COREAUDIO_MIDI)
	midi_fd = -1;	/* CoreAudio notifications stand in for select here */
	/* XXX: here we initialize CoreAudio midi listener... see
	 * /Developer/Examples/CoreAudio/MIDI/SampleTools/Echo.cpp
	 * and then add the callback function... */
#endif
	return 0;
}

void
midi_in(void)
{
	unsigned char buf[32];
	int i,n;
	static enum {
		m_start,
		m_note,
		m_note_key,
		m_note_off,
		m_note_off_key,
		m_sysex1,
		m_sysex2,
		m_sysex3,
		m_sysex4,
		m_sysex5,
		m_sysex6
	} state = m_start;
	static unsigned char note = 0;

	n = read(midi_fd,buf,sizeof buf);
	if (n == 0) {
		close(midi_fd);
		midi_fd = -1;
		fprintf(stderr,"EOF on MIDI!\n");
		return;
	}
	if (n < 0) {
		perror("read(midi)");
		return;
	}
	for (i = 0; i < n; i++) {
		unsigned char c = buf[i];
		if (c & 0x80) {
			/* command */
			if (c == 0x80) {
				state = m_note_off;
			} else if (c == 0x90) {
				state = m_note;
			} else if (c == 0xF0) {
				state = m_sysex1;
			} else if (c == 0xF7) {
				/* just ignore */
				state = m_start;
			} else {
				/* unknown command */
				state = m_start;
				printf("MIDI command %02X\n",(int)c);
			}
		} else if (state == m_note) {
			note = c;
			state = m_note_key;
		} else if (state == m_note_key) {
			midi_record_event('N',note,c);
			cmd_note(note,c);
			state = m_note;
		} else if (state == m_note_off) {
			note = c;
			state = m_note_off_key;
		} else if (state == m_note_off_key) {
			midi_record_event('U',note,0);
			cmd_note(note,0);
			state = m_note_off;
		} else if (state == m_sysex1 && c == 0x18) {
			state = m_sysex2;
		} else if (state == m_sysex2 && c == 0x41) {
			state = m_sysex3;
		} else if (state == m_sysex3 && c == 0x00) {
			state = m_sysex4;
		} else if (state == m_sysex4 && c == 0x00) {
			state = m_sysex5;
		} else if (state == m_sysex5) {
			note = c;
			state = m_sysex6;
		} else if (state == m_sysex6) {
			if (c) {
				/* note up */
			} else {
				midi_record_event('S',note,0);
				cmd_sysex(note);
			}
			state = m_start;
		} else {
			printf("MIDI byte %02X (in state %d)\n",(int)c,state);
			state = m_start;
		}
	}
}

void
midi_cleanup(void)
{
	close(midi_fd);
}



static FILE *midi_record_f = NULL;
static struct timeval midi_record_tv = { 0, 0 };
static unsigned long midi_record_count = 0;		/* just for comments */

static void
midi_record_event(char kind, int x, int y)
{
	struct timeval tv;
	char comment[80];

	if (!midi_record_f) return;

	if (midi_record_tv.tv_sec == 0) {
		gettimeofday(&midi_record_tv, NULL);
	} else {
		unsigned elapsed_ms;

		gettimeofday(&tv, NULL);

		elapsed_ms = (tv.tv_sec - midi_record_tv.tv_sec) * 1000;
		elapsed_ms += ((tv.tv_usec/1000)+1000) - (midi_record_tv.tv_usec/1000);
		elapsed_ms -= 1000;
		if (elapsed_ms >= 1) {
			fprintf(midi_record_f, "T%u\n",elapsed_ms);
			midi_record_tv = tv;
		}
		midi_record_count += elapsed_ms;
	}

	strcpy(comment,"# ");
	if (kind == 'N' || kind == 'U') {
		assert_str(x >= 0 && x <= 127, "recording invalid midi event?");
		sprintf(comment+strlen(comment), "%s%s%s ",
			((kind == 'U') ? "(" : ""), midi_notetab[x],
			((kind == 'U') ? ")" : ""));
	}
	sprintf(comment+strlen(comment), "@ %lu", midi_record_count);

	if (y) {
		fprintf(midi_record_f, "%c%d,%d\t%s\n", kind, x, y, comment);
	} else if (x) {
		fprintf(midi_record_f, "%c%d\t%s\n", kind, x, comment);
	} else {
		fprintf(midi_record_f, "%c\t%s\n", kind, comment);
	}
}

void
midi_record_cmd(char *arg)
{
	char *fn;

	arg += strspn(arg," \t");
	arg[strcspn(arg," \t")] = 0;
	fn = extended_fn(arg, "evt");

	midi_stop_cmd();

	midi_record_f = fopen(fn, "w");
	if (!midi_record_f) {
		perror(fn);
		return;
	}
	midi_record_tv.tv_sec = 0;
}


static FILE *midi_play_f = NULL;
static struct timeval midi_play_tv = { 0, 0 };

void
midi_play_step(void)
{
	struct timeval tv;
	char buf[40];
	char kind;
	int x = 0, y = 0;
	int i;

	if (!midi_play_f) return;

	gettimeofday(&tv, NULL);

	if ((tv.tv_sec < midi_play_tv.tv_sec) ||
	    ((tv.tv_sec == midi_play_tv.tv_sec) &&
	     (tv.tv_usec < midi_play_tv.tv_usec))) {
		return;
	}

	if (midi_play_tv.tv_sec == 0) {
		midi_play_tv = tv;
	}

	if (!fgets(buf, 40, midi_play_f)) {
		/* done with file */
		fclose(midi_play_f);
		midi_play_f = NULL;
		printf("\rmidi playback completed\n");
		return;
	}
	buf[strcspn(buf," \t\r\n")] = 0;
	if (!buf[0]) {
		/* try again - next line */
		midi_play_step();
		return;
	}

	kind = buf[0];
	i = 1;
	while (isdigit(buf[i])) {
		x = (x*10) + (buf[i]-'0');
		i++;
	}
	if (buf[i] == ',') {
		i++;
		while (isdigit(buf[i])) {
			y = (y*10) + (buf[i]-'0');
			i++;
		}
	}

	switch(kind) {
		case 'T': /* update delay and then try next line */
			midi_play_tv.tv_usec += x*1000;
			midi_play_tv.tv_sec += midi_play_tv.tv_usec/1000000;
			midi_play_tv.tv_usec %= 1000000;
			midi_play_step();
			break;
		case 'N':
			cmd_note(x, y);
			break;
		case 'U':
			cmd_note(x, 0);
			break;
		case 'S':
			cmd_sysex(x);
			break;
		default:	/* silently ignore gibberish */
			break;
	}
}

void
midi_play_cmd(char *arg)
{
	char *fn;

	arg += strspn(arg," \t");
	arg[strcspn(arg," \t")] = 0;
	fn = extended_fn(arg, "evt");

	midi_stop_cmd();

	midi_play_f = open_search(fn, "r");
	if (!midi_play_f) {
		perror(fn);
		return;
	}
	midi_play_tv.tv_sec = 0;
}


void
midi_stop_cmd()
{
	if (midi_record_f) {
		printf("ending recording...\n");
		fclose(midi_record_f);
		midi_record_f = NULL;
	}
	if (midi_play_f) {
		printf("ending playback...\n");
		fclose(midi_play_f);
		midi_play_f = NULL;
	}
}


void
midi_status_cmd(void)
{
	if (midi_record_f) {
		printf("currently recording MIDI events...\n");
	}
	if (midi_play_f) {
		printf("currently playing back MIDI events...\n");
	}
}
