/* Copyright 2010 Greg Alexander.  See the file COPYING for terms */
/* $Id: opt.c,v 1.9 2010-02-08 03:17:28 greg Exp $ */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "lang.h"
#include "misc.h"

/* This optimizer takes a function definition tree and fills in its opt_expr
 * member with a flattened version (no op_fn_calls) of the expr member,
 * much like one_tree() in eval.c (but it doesn't allocate temps and it
 * doesn't generate an op_one_tree).  Then several optimizations are performed
 * on the flattened expression:
 *  - constant folding (and associative laws)
 *  - raising expressions that depend only on formals (computed once per note)
 *  - common subexpressions (for formal expansion duplicates)
 */


struct formals_rec {
	struct formals_rec *context;
	struct tree_rec **formals;
	int nformals;
};

static struct tree_rec *opt_one_tree(struct tree_rec *tr, struct formals_rec *formals);
static void opt_constant_fold(struct tree_rec *tr);
static void opt_constant_fold_arith(struct tree_rec *tr);
static void opt_raise_formal_constants(struct tree_rec *tr);
static void opt_raise_common_exprs(struct tree_rec *tr);

static int changed = 0;


void
opt(struct tree_rec *tr)
{
	assert_str(tr->kind == op_fn_def, "opt() on non-function?");
	assert_str((tr->u.fn_def.expr != NULL), "opt() with no function body?");

	tr->u.fn_def.opt_expr = opt_one_tree(tr->u.fn_def.expr, NULL);
	if (debug_flags&1) {
		printf("opt...one tree: ");
		dump_tree(tr->u.fn_def.opt_expr);
		printf("\n");
	}
	do {
		changed = 0;
		opt_constant_fold(tr->u.fn_def.opt_expr);
		if (changed && (debug_flags&1)) {
			printf("opt...folded: ");
			dump_tree(tr->u.fn_def.opt_expr);
			printf("\n");
		}
	} while (changed);

	opt_raise_formal_constants(tr);
	if (debug_flags&1) {
		printf("opt...raised formals:\n");
		dump_tree(tr->u.fn_def.opt_expr);
		printf("\n");
	}

	opt_raise_common_exprs(tr->u.fn_def.opt_expr);
	if (debug_flags&1) {
		printf("opt..raised common exprs:\n");
		dump_tree(tr->u.fn_def.opt_expr);
		printf("\n");
	}
}

static struct tree_rec *
opt_one_tree(struct tree_rec *tr, struct formals_rec *formals)
{
	struct tree_rec *ret = NULL;

	switch (tr->kind) {
		case op_error:		/* these are all illegal here */
		case op_fn_def:
		case op_stereo:
		case op_one_tree:
		case op_raised:
			do_error("unhandled tree in opt_one_tree()");
			ret = new_tree(op_error);
			break;
		case op_fn_call: {
			struct formals_rec *nested;
			struct tree_rec *fn_def = tr->u.fn_call.func->defn;

			if (!fn_def) {
				/* the function is undefined (maybe builtin?)
				 * leave it for later, but flatten its parms */
				struct builtin_rec *bi;
				int n = tr->u.fn_call.num_args;
				int i;

				ret = new_tree(op_fn_call);
				ret->u.fn_call.func = tr->u.fn_call.func;
				ret->u.fn_call.args = safe_malloc(n*sizeof *ret->u.fn_call.args);
				ret->u.fn_call.num_args = n;
				/* don't bother with num_temps/num_itemps */
				for (i = 0; i < n; i++) {
					ret->u.fn_call.args[i] = opt_one_tree(tr->u.fn_call.args[i], formals);
				}
				bi = tr->u.fn_call.func->builtin;
				if (bi && (bi->num_args >= 1) &&
				    (bi->wave&WAV_FREQ)) {
					struct tree_rec *arg0 = ret->u.fn_call.args[0];
					if (arg0->kind == op_timest &&
					    arg0->u.timest.kind == t_default) {
						arg0->u.timest.kind = t_freq;
					}
				}
				break;
			}
			if (tr->u.fn_call.num_args != fn_def->u.fn_def.num_formals) {
				do_error("function %s called with wrong number of args",tr->u.fn_call.func->name);
				return new_tree(op_error);
			}
			nested = (struct formals_rec *)safe_malloc(sizeof *nested);
			nested->formals = tr->u.fn_call.args;
			nested->nformals = tr->u.fn_call.num_args;
			nested->context = formals;
			ret = opt_one_tree(fn_def->u.fn_def.expr, nested);
			free(nested);
		}	break;
		case op_add:
		case op_subtract:
		case op_multiply:
		case op_divide:
			ret = new_tree(tr->kind);
			ret->u.operands[0] = opt_one_tree(tr->u.operands[0], formals);
			ret->u.operands[1] = opt_one_tree(tr->u.operands[1], formals);
			break;
		case op_termseq: {
			int n = tr->u.termseq.num_intervals;
			int i;

			ret = new_tree(op_termseq);
			ret->u.termseq.durations = safe_malloc(n*sizeof *ret->u.termseq.durations);
			ret->u.termseq.exprs = safe_malloc(n*sizeof *ret->u.termseq.exprs);
			ret->u.termseq.zero_crossing = safe_malloc(n*sizeof *ret->u.termseq.zero_crossing);
			memcpy(ret->u.termseq.zero_crossing, tr->u.termseq.zero_crossing, n*sizeof *ret->u.termseq.zero_crossing);
			ret->u.termseq.num_intervals = n;
			ret->u.termseq.repeats = tr->u.termseq.repeats;
			for (i = 0; i < n; i++) {
				ret->u.termseq.durations[i] = opt_one_tree(tr->u.termseq.durations[i], formals);
				ret->u.termseq.exprs[i] = opt_one_tree(tr->u.termseq.exprs[i], formals);
			}
		}	break;
		case op_number:
		case op_string:
			ret = dup_tree(tr);
			break;
		case op_argref:
			if (!formals) {
				/* top-level argref, pass through */
				ret = dup_tree(tr);
			} else if (tr->u.argref >= formals->nformals) {
				do_error("reference past end of formals in opt_one_tree()");
				ret = new_tree(op_error);
			} else {
				ret = opt_one_tree(formals->formals[tr->u.argref], formals->context);
			}
			break;
		case op_timest:
			ret = new_tree(op_timest);
			ret->u.timest.operand = opt_one_tree(tr->u.timest.operand, formals);
			ret->u.timest.kind = tr->u.timest.kind;
			break;
	}
	return ret;
}


#define IS_CONST(tr) ((tr)->kind == op_number)
#define IS_SUM(tr) (((tr)->kind == op_add) || ((tr)->kind == op_subtract))
#define IS_PROD(tr) (((tr)->kind == op_multiply) || ((tr)->kind == op_divide))
#define VAL(tr) ((tr)->u.number)

/* this applies a number of rules to put tr into a canonical organization
 * with all of the constants that can be folded together.
 * It performs a depth-first search.  If it makes any changes then it sets the
 * global variable "changed", so it can be repeated.
 */
static void
opt_constant_fold(struct tree_rec *tr)
{
	int is_arith = 0;
		/* first, fold the subtrees */
	switch (tr->kind) {
		case op_error:		/* these are all illegal here */
		case op_fn_def:
		case op_stereo:
		case op_one_tree:
		case op_raised:
			do_error("unhandled tree in opt_constant_fold()");
			break;
		case op_fn_call: {	/* builtin function calls can be here */
			int i;
			for (i = 0; i < tr->u.fn_call.num_args; i++) {
				opt_constant_fold(tr->u.fn_call.args[i]);
			}
		}	break;
		case op_add:
		case op_subtract:
		case op_multiply:
		case op_divide:
			opt_constant_fold(tr->u.operands[0]);
			opt_constant_fold(tr->u.operands[1]);
			is_arith = 1;
			break;
		case op_termseq: {
			int n = tr->u.termseq.num_intervals;
			int i;
			for (i = 0; i < n; i++) {
				opt_constant_fold(tr->u.termseq.durations[i]);
				opt_constant_fold(tr->u.termseq.exprs[i]);
			}
		}	break;
		case op_number:
		case op_string:
			/* already a constant, nothing doing */
			break;
		case op_argref:
			/* must be a top-level formal, nothing doing */
			break;
		case op_timest:
			opt_constant_fold(tr->u.timest.operand);
			break;
	}
	if (is_arith) {
		opt_constant_fold_arith(tr);
	} else if ((tr->kind == op_fn_call) &&
		   (tr->u.fn_call.num_args == 1) &&
		   (IS_CONST(tr->u.fn_call.args[0]))) {
/* RULE: builtin(const) --> const */
		struct tree_rec *x = tr->u.fn_call.args[0];
		free(tr->u.fn_call.args);
		if (!strcmp(tr->u.fn_call.func->name, "sin")) {
			VAL(tr) = sin(VAL(x));
		} else if (!strcmp(tr->u.fn_call.func->name, "cos")) {
			VAL(tr) = cos(VAL(x));
		} else if (!strcmp(tr->u.fn_call.func->name, "exp")) {
			VAL(tr) = exp(VAL(x));
		} else {
			do_error("unknown builtin clobbered by opt_constant_fold");
		}
		tr->kind = op_number;
		free_tree(x);
	}

}

/* implement the arithmetic operator part of constant folding */
static void
opt_constant_fold_arith(struct tree_rec *tr)
{
	struct tree_rec *lhs = tr->u.operands[0];
	struct tree_rec *rhs = tr->u.operands[1];
	tree_op op = tr->kind;
	tree_op lhs_op = lhs->kind;
	tree_op rhs_op = rhs->kind;
	struct tree_rec *orig = 0;

	if (debug_flags&1) {
		orig = dup_tree(tr);
	}

	if (IS_CONST(lhs) && IS_CONST(rhs)) {
/* RULE: const OP const --> const          (OP is of + - * /) */
		switch(op) {
			case op_add:      VAL(tr) = VAL(lhs) + VAL(rhs); break;
			case op_subtract: VAL(tr) = VAL(lhs) - VAL(rhs); break;
			case op_multiply: VAL(tr) = VAL(lhs) * VAL(rhs); break;
			case op_divide:   VAL(tr) = VAL(lhs) / VAL(rhs); break;
			default: break;	/* no warning */
		}
		tr->kind = op_number;
		free_tree(lhs);
		free_tree(rhs);
		goto success;
	} 
	if (IS_CONST(rhs)) {
/* RULE: x + const      --> const + x      constants on LHS */
/* RULE: x - const      --> (-const) + x   normalize to addition */
/* RULE: x * const      --> const * x      constants on LHS */
/* RULE: x / const      --> (1/const) * x  normalize to multiplication */
		tr->u.operands[0] = rhs;
		tr->u.operands[1] = lhs;
		if (op == op_subtract) {
			VAL(rhs) *= -1.0;
			tr->kind = op_add;
		} else if (op == op_divide) {
			VAL(rhs) = 1.0 / VAL(rhs);
			tr->kind = op_multiply;
		}
		goto success;
	}

	if (((op == op_add) && IS_CONST(lhs) &&
	     IS_SUM(rhs) && IS_CONST(rhs->u.operands[0])) ||
	    ((op == op_multiply) && IS_CONST(lhs) &&
	     IS_PROD(rhs) && IS_CONST(rhs->u.operands[0]))) {
/* RULE: c1 + (c2 OP x) --> (c1 + c2) OP x associative rule (OP is of + -) */
/* RULE: c1 * (c2 OP x) --> (c1 * c2) OP x associative rule (OP is of * /) */
		struct tree_rec *c2 = rhs->u.operands[0];
		struct tree_rec *x = rhs->u.operands[1];
		tr->u.operands[0] = rhs;
		rhs->u.operands[0] = lhs;
		rhs->u.operands[1] = c2;
		rhs->kind = op;
		tr->u.operands[1] = x;
		tr->kind = rhs_op;
		goto success;
	}
	if ((IS_SUM(tr) && IS_SUM(lhs)) ||
	    (IS_PROD(tr) && IS_PROD(lhs))) {
/* RULE: (x OP y) OP z  --> x OP (y OP z)  associative rule (OP is of + -) */
/* RULE: (x OP y) OP z  --> x OP (y OP z)  associative rule (OP is of * /) */
		struct tree_rec *x = lhs->u.operands[0];
		struct tree_rec *y = lhs->u.operands[1];
		struct tree_rec *z = rhs;
		tree_op op1 = lhs_op;
		tree_op op2 = op;
		tr->u.operands[0] = x;
		tr->u.operands[1] = lhs;
		lhs->u.operands[0] = y;
		lhs->u.operands[1] = z;
		if ((op1 == op_add) || (op1 == op_multiply)) {
/* (x+y)+z --> x+(y+z)
 * (x+y)-z --> x+(y-z)
 * (x*y)*z --> x*(y*z)
 * (x*y)/z --> x*(y/z)
			 */
			tr->kind = op1;
			lhs->kind = op2;
		} else if ((op2 == op_add) || (op2 == op_multiply)) {
/* (x-y)+z --> x+(z-y)
 * (x/y)*z --> x*(z/y)
 */
			tr->kind = op2;
			lhs->kind = op1;
			lhs->u.operands[0] = z;
			lhs->u.operands[1] = y;
		} else {
/* (x-y)-z --> x-(y+z)
 * (x/y)/z --> x/(y*z)
 */
			tr->kind = op1;
			lhs->kind = ((op1 == op_subtract) ? op_add : op_multiply);
		}
		goto success;
	}
	if (((IS_SUM(tr) && IS_SUM(rhs)) ||
	     (IS_PROD(tr) && IS_PROD(rhs))) &&
	    IS_CONST(rhs->u.operands[0])) {
/* RULE: x OP (c OP y)  --> c OP (x OP y)  associative rule (OP is of + -) */
/* RULE: x OP (c OP y)  --> c OP (x OP y)  associative rule (OP is of * /) */
		struct tree_rec *x = lhs;
		struct tree_rec *c = rhs->u.operands[0];
		struct tree_rec *y = rhs->u.operands[1];
		tree_op op1 = op;
		tree_op op2 = rhs_op;
		tr->u.operands[0] = c;
		tr->u.operands[1] = rhs;
		rhs->u.operands[0] = x;
		rhs->u.operands[1] = y;
		if ((op1 == op_add) || (op1 == op_multiply)) {
/* x+(c+y) --> c+(x+y) */
/* x+(c-y) --> c+(x-y) */
/* x*(c*y) --> c*(x*y) */
/* x*(c/y) --> c*(x/y) */
			tr->kind = op1;
			rhs->kind = op2;
		} else if ((op2 == op_add) || (op2 == op_multiply)) {
/* x-(c+y) --> -c+(x-y) */
/* x/(c*y) --> (1/c)*(x/y) */
			tr->kind = op2;
			rhs->kind = op1;
			if (op2 == op_add) {
				VAL(c) *= -1.0;
			} else {
				VAL(c) = 1.0 / VAL(c);
			}
		} else {
/* x-(c-y) --> -c+(x+y) */
/* x/(c/y) --> (1/c)*(x*y) */
			tr->kind = rhs->kind =
				((op1 == op_subtract) ? op_add : op_multiply);
			if (op1 == op_subtract) {
				VAL(c) *= -1.0;
			} else {
				VAL(c) = 1.0 / VAL(c);
			}
		}
		goto success;
	}
	if ((op == op_divide) && tree_equal(lhs,rhs)) {
/* RULE: x / x          --> 1 */
		tr->kind = op_number;
		VAL(tr) = 1.0;
		free_tree(lhs);
		free_tree(rhs);
		goto success;
	}
	if ((op == op_subtract) && tree_equal(lhs,rhs)) {
/* RULE: x - x          --> 0 */
		tr->kind = op_number;
		VAL(tr) = 0.0;
		free_tree(lhs);
		free_tree(rhs);
		goto success;
	}
	if (IS_CONST(lhs) && (VAL(lhs) == 0.0) && (op == op_add)) {
/* RULE: 0 + x          --> x */
		memcpy(tr, rhs, sizeof *tr);
		free_tree(lhs);
		free(rhs);
		goto success;
	}
	if (IS_CONST(lhs) && (VAL(lhs) == 1.0) && (op == op_multiply)) {
/* RULE: 1 * x          --> x */
		memcpy(tr, rhs, sizeof *tr);
		free_tree(lhs);
		free(rhs);
		goto success;
	}
	if (IS_CONST(lhs) && (VAL(lhs) == 0.0) && (op == op_multiply)) {
/* RULE: 0 * x          --> 0 */
		tr->kind = op_number;
		VAL(tr) = 0.0;
		free_tree(lhs);
		free_tree(rhs);
		goto success;
	}

	if (debug_flags&1) {
		free_tree(orig);
	}
	return;
success:
	changed++;
	if (debug_flags&1) {
		printf("CHANGED: ");
		dump_tree(orig);
		printf(" --> ");
		dump_tree(tr);
		printf("\n");
		free_tree(orig);
	}
}

static struct tree_rec **raised = NULL;
static int num_raised = 0;
static int size_raised = 0;

static int
non_trivial(struct tree_rec *tr)
{
	switch (tr->kind) {
		case op_fn_call:
		case op_add:
		case op_subtract:
		case op_multiply:
		case op_divide:
		case op_termseq:
		case op_timest:
			return 1;
		default:
			return 0;
	}
}

static int
uses_t(struct tree_rec *tr)
{
	int i;

	switch (tr->kind) {
		case op_error:
		case op_fn_def:
		case op_stereo:
		case op_one_tree:
		case op_raised:
			/* shrug */
			return 1;
		case op_fn_call:
			if (!strcmp(tr->u.fn_call.func->name, "white")) {
				return 1;
			}
			for (i = 0; i < tr->u.fn_call.num_args; i++) {
				if (uses_t(tr->u.fn_call.args[i])) {
					return 1;
				}
			}
			return 0;
		case op_add:
		case op_subtract:
		case op_multiply:
		case op_divide:
			return (uses_t(tr->u.operands[0]) ||
			        uses_t(tr->u.operands[1]));
		case op_termseq:
			return 1;
		case op_number:
		case op_argref:
		case op_string:
			return 0;
		case op_timest:
			return 1;
	}
	return 1;
}

static void
add_raised(struct tree_rec *tr)
{
	if (num_raised >= size_raised) {
		struct tree_rec **n;
		int nlen = size_raised+10;
		n = malloc(nlen*sizeof *n);
		memcpy(n, raised, num_raised*sizeof *n);
		free(raised);
		raised = n;
		size_raised = nlen;
	}
	raised[num_raised] = tr;
	num_raised++;
}

static void
raise_formal_cons(struct tree_rec **trp)
{
	struct tree_rec *tr = *trp;
	if (non_trivial(tr) && !uses_t(tr)) {
		struct tree_rec *ref;
		ref = new_tree(op_argref);
		ref->u.argref = num_raised|ARG_RAISED_FORMAL;
		add_raised(tr);
		*trp = ref;
	} else {
		/* search children for raisable values */
		switch (tr->kind) {
			case op_error:
			case op_fn_def:
			case op_one_tree:
			case op_raised:
			case op_number:
			case op_argref:
			case op_string:
			case op_stereo:
				/* whatever */
				break;
			case op_fn_call: {
				int i;
				for (i = 0; i < tr->u.fn_call.num_args; i++) {
					raise_formal_cons(&tr->u.fn_call.args[i]);
				}
			}	break;
			case op_add:
			case op_subtract:
			case op_multiply:
			case op_divide:
				raise_formal_cons(&tr->u.operands[0]);
				raise_formal_cons(&tr->u.operands[1]);
				break;
			case op_termseq: {
				int i;
				for (i = 0; i < tr->u.termseq.num_intervals; i++) {
					raise_formal_cons(&tr->u.termseq.durations[i]);
					raise_formal_cons(&tr->u.termseq.exprs[i]);
				}
			}	break;
			case op_timest:
				raise_formal_cons(&tr->u.timest.operand);
				break;
		}
	}
}

static void
opt_raise_formal_constants(struct tree_rec *tr)
{
	struct tree_rec *defn;
	struct tree_rec *r;

	defn = tr->u.fn_def.opt_expr;
	raise_formal_cons(&defn);

	r = new_tree(op_raised);
	r->u.raised.expr = defn;
	r->u.raised.formals = raised;
	r->u.raised.num_formals = num_raised;
	tr->u.fn_def.opt_expr = r;

	raised = NULL;
	num_raised = size_raised = 0;
}


struct common_expr_rec {
	struct common_expr_rec *next;
	struct tree_rec *expr;
	struct tree_rec **ref;		/* where first reference was from */
	int index;			/* -1 if not risen yet */
};

static struct common_expr_rec *common_exprs = NULL;

static struct common_expr_rec *
already_common(struct tree_rec *tr)
{
	struct common_expr_rec *c;
	for (c = common_exprs; c; c = c->next) {
		if (tree_equal(c->expr, tr)) {
			return c;
		}
	}
	return NULL;
}

static void
remove_common(struct tree_rec *tr)
{
	struct common_expr_rec **p;
	for (p = &common_exprs; *p; p = &((*p)->next)) {
		if ((*p)->expr == tr) {
			struct common_expr_rec *t = *p;
			*p = t->next;
			free(t);
		}
	}
}

/* removes all subexpressions of tr from the common_exprs list, because tr is
 * being raised to a common subexpression and that should not happen
 * piecemeal, so none of its children can now be raised... */
static void
remove_subexprs(struct tree_rec *tr)
{
	int i;
	switch (tr->kind) {
		case op_error:
		case op_fn_def:
		case op_stereo:
		case op_one_tree:
		case op_raised:
			/* whatev */
			break;
		case op_fn_call:
			for (i = 0; i < tr->u.fn_call.num_args; i++) {
				remove_common(tr->u.fn_call.args[i]);
				remove_subexprs(tr->u.fn_call.args[i]);
			}
		case op_add:
		case op_subtract:
		case op_multiply:
		case op_divide:
			remove_common(tr->u.operands[0]);
			remove_subexprs(tr->u.operands[0]);
			remove_common(tr->u.operands[1]);
			remove_subexprs(tr->u.operands[1]);
			break;
		case op_termseq:
			for (i = 0; i < tr->u.termseq.num_intervals; i++) {
				remove_common(tr->u.termseq.durations[i]);
				remove_subexprs(tr->u.termseq.durations[i]);
				remove_common(tr->u.termseq.exprs[i]);
				remove_subexprs(tr->u.termseq.exprs[i]);
			}
			break;
		case op_number:
		case op_argref:
		case op_string:
			/* no children */
			break;
		case op_timest:
			remove_common(tr->u.timest.operand);
			remove_subexprs(tr->u.timest.operand);
			break;
	}
}

static void
raise_commons(struct tree_rec **trp)
{
	struct tree_rec *tr = *trp;
	int i;

	/* see if tr is common */
	if (non_trivial(tr)) {
		struct common_expr_rec *c;
		struct tree_rec *n;
		if ((c=already_common(tr))) {
			if (c->index == -1) {
				c->index = num_raised;
				add_raised(c->expr);
				n = new_tree(op_argref);
				n->u.argref = c->index | ARG_RAISED_COMMON;
				(*(c->ref)) = n;
				remove_subexprs(c->expr);
			}
			n = new_tree(op_argref);
			n->u.argref = c->index | ARG_RAISED_COMMON;
			*trp = n;
			free_tree(tr);
		} else {
			/* not already present, add it */
			c = malloc(sizeof *c);
			c->next = common_exprs;
			c->expr = tr;
			c->ref = trp;
			c->index = -1;
			common_exprs = c;
		}
	}

	/* recur on the children */
	switch (tr->kind) {
		case op_error:
		case op_fn_def:
		case op_stereo:
		case op_one_tree:
		case op_raised:
			/* whatev */
			break;
		case op_fn_call:
			for (i = 0; i < tr->u.fn_call.num_args; i++) {
				raise_commons(&tr->u.fn_call.args[i]);
			}
			break;
		case op_add:
		case op_subtract:
		case op_multiply:
		case op_divide:
			raise_commons(&tr->u.operands[0]);
			raise_commons(&tr->u.operands[1]);
			break;
		case op_termseq:
			for (i = 0; i < tr->u.termseq.num_intervals; i++) {
				raise_commons(&tr->u.termseq.durations[i]);
				raise_commons(&tr->u.termseq.exprs[i]);
			}
			break;
		case op_number:
		case op_argref:
		case op_string:
			/* no children */
			break;
		case op_timest:
			raise_commons(&tr->u.timest.operand);
			break;
	}
}

static void
opt_raise_common_exprs(struct tree_rec *tr)
{
	struct common_expr_rec *ce, *ce_next;

	assert_str(tr->kind == op_raised, "raise_common_exprs must already have raised op");
	raise_commons(&tr->u.raised.expr);
	tr->u.raised.commons = raised;
	tr->u.raised.num_commons = num_raised;

	raised = NULL;
	num_raised = size_raised = 0;

	for (ce = common_exprs; ce; ce = ce_next) {
		ce_next = ce->next;
		free(ce);
	}
	common_exprs = NULL;
}
