/* Copyright 2010 Greg Alexander.  See the file COPYING for terms */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/time.h>
#include "misc.h"
#include "snd.h"
#include "midi.h"

/* at the top level, next is a pointer to other patterns, and stack_next
 * is a pointer to a stack of old versions of this pattern that we can
 * revert to.  within that stack, stack_next points to the next element
 * in the stack and next should be NULL. */
struct pattern_rec {
	struct pattern_rec *next;
	struct pattern_rec *stack_next;
	char *cmd;			/* command that made this stack entry */
	int patno;
	struct pattern_note_rec *notes;
	int num_measures;
	unsigned char muted;		/* ==0 means playing */
	char inst[40];			/* the function for this pattern */
};

struct pattern_note_rec {
	struct pattern_note_rec *next;
	int ofs;			/* offset in ms from start of pattern */
	unsigned char note;		/* midi note value (0-127) */
	unsigned char vel;		/* midi velocity (0-127) */
	unsigned duration;		/* length in ms */
	char *inst;			/* just a pointer to pattern_rec.inst */
	int patno;
};

static void pattern_tempo(int argc, char **argv);
static void pattern_new(int argc, char **argv);
static void pattern_mute(void);
static void pattern_start(void);
static void pattern_write(int argc, char **argv);
static void pattern_shift(int argc, char **argv);
static void pattern_inst(int argc, char **argv);
static void pattern_measures(int argc, char **argv);
static void pattern_rhythm(int argc, char **argv);
static void pattern_velocity(int argc, char **argv);
static void pattern_revert(int argc, char **argv);
static void pattern_save_cmd(int argc, char **argv);
static void pattern_list(int argc, char **argv);
static void pattern_delete(int argc, char **argv);
static void pattern_edit(int argc, char **argv);
static void pattern_note_list(struct pattern_note_rec *pn, int measures);
static void pattern_copy(struct pattern_rec *dst, struct pattern_rec *src);
static void pattern_save(char *cmd);
static void pattern_trace(int argc, char **argv);
static void pattern_reset(void);
static void delete_pattern(struct pattern_rec *pat);


static struct pattern_rec *patterns = NULL;
static struct pattern_rec *patterns_last = NULL;
static struct pattern_rec *curr_pattern = NULL;
static int curr_pattern_explicit = 0;
static int curr_tempo = 10000;		/* default to 1 second */
	/* pending_notes is a queue in increasing order.  Elements are added
	 * where they belong, then removed at the beginning by pattern_step()
	 */
static struct pattern_note_rec *pending_notes = NULL;
	/* curr_offset is the global pattern clock counting from "pattern
	 * start", used for pending_notes offsets */
static int curr_offset = 0;
static int playback_going = 0;
static int playback_trace = 0;
static int playback_trace_beats = 0;
	/* rec_pat is the pattern we are recording into...if it is set,
	 * but record_going is not, it means we are waiting for a measure
	 * boundary to sync up. */
static struct pattern_rec *rec_pat = NULL;
static struct pattern_note_rec *rec_pat_note_last = NULL;
static struct timeval rec_start_time = { 0,0 };
static int record_going = 0;
static int determine_tempo = 0;


void
pattern_cmd(char *args)
{
	static char **argv = NULL;
	int argc = 0;
	static int max_argc = 0;

			/* handle the optional pattern selector */
	args += strspn(args, " \t");
	if (isdigit(*args)) {
		int patno = strtoul(args, &args, 10);
		struct pattern_rec *p;
		for (p = patterns; p; p = p->next) {
			if (p->patno == patno) {
				break;
			}
		}
		if (p) {
			curr_pattern = p;
			curr_pattern_explicit = 1;
		} else {
			fprintf(stderr, "couldn't find pattern %d, ignoring pattern specification.\n", patno);
			curr_pattern_explicit = 0;
		}
		args += strspn(args, " \t");
	} else {
		curr_pattern_explicit = 0;
	}

			/* split it into argv/argc */
	while (*args) {
		char *t;
		t = args + strcspn(args, " \t");
		if (*t) {
			*t = 0;
			t++;
		}
		if (!*args) break;
		if (argc+2 >= max_argc) {
			int t_argc = argc + 10;
			char **t_argv = (char **)safe_malloc(t_argc*sizeof *t_argv);
			memset(t_argv, 0, t_argc*sizeof *t_argv);
			if (argc) {
				memcpy(t_argv, argv, argc*sizeof *t_argv);
			}
			free(argv);
			max_argc = t_argc;
			argv = t_argv;
		}
		argv[argc++] = args;
		args = t + strspn(t, " \t");
	}

	if (argc < 1) {
		fprintf(stderr, "pattern subcommand not specified.  help:\n");
		help("pattern");
		return;
	}

	if (!strcasecmp(argv[0], "tempo")) {
		pattern_tempo(argc, argv);
	} else if (!strcasecmp(argv[0], "new")) {
		pattern_new(argc, argv);
	} else if (!strcasecmp(argv[0], "mute")) {
		pattern_mute();
	} else if (!strcasecmp(argv[0], "start")) {
		pattern_start();
	} else if (!strcasecmp(argv[0], "stop")) {
		pattern_stop();
	} else if (!strcasecmp(argv[0], "write")) {
		pattern_write(argc, argv);
	} else if (!strcasecmp(argv[0], "shift")) {
		pattern_shift(argc, argv);
	} else if (!strcasecmp(argv[0], "inst")) {
		pattern_inst(argc, argv);
	} else if (!strcasecmp(argv[0], "measures")) {
		pattern_measures(argc, argv);
	} else if (!strcasecmp(argv[0], "rhythm")) {
		pattern_rhythm(argc, argv);
	} else if (!strcasecmp(argv[0], "velocity")) {
		pattern_velocity(argc, argv);
	} else if (!strcasecmp(argv[0], "revert")) {
		pattern_revert(argc, argv);
	} else if (!strcasecmp(argv[0], "save")) {
		pattern_save_cmd(argc, argv);
	} else if (!strcasecmp(argv[0], "list")) {
		pattern_list(argc, argv);
	} else if (!strcasecmp(argv[0], "delete")) {
		pattern_delete(argc, argv);
	} else if (!strcasecmp(argv[0], "edit")) {
		pattern_edit(argc, argv);
	} else if (!strcasecmp(argv[0], "trace")) {
		pattern_trace(argc, argv);
	} else if (!strcasecmp(argv[0], "reset")) {
		pattern_reset();
	} else {
		fprintf(stderr, "pattern subcommand \"%s\" not found.  help:\n", argv[0]);
		help("pattern");
	}
}


static struct pattern_note_rec *pattern_note_pool = NULL;

static struct pattern_note_rec *
pattern_note_giveme(void)
{
	struct pattern_note_rec *ret;
	if (pattern_note_pool) {
		ret = pattern_note_pool;
		pattern_note_pool = ret->next;
	} else {
		ret = safe_malloc(sizeof *ret);
	}
	memset(ret, 0, sizeof *ret);
	return ret;
}

static void
pattern_note_takeit(struct pattern_note_rec *pn)
{
	pn->next = pattern_note_pool;
	pattern_note_pool = pn;
}

static void
pattern_note_list_takeit(struct pattern_note_rec *pn)
{
	struct pattern_note_rec *pn_next;

			/* free all pending notes.. */
	for (; pn; pn = pn_next) {
		pn_next = pn->next;
		pattern_note_takeit(pn);
	}
}


/* copies the notes for a pattern into the ordered pending notes queue */
static void
pattern_to_pending(struct pattern_rec *p)
{
	struct pattern_note_rec *pn;

	for (pn = p->notes; pn; pn = pn->next) {
		struct pattern_note_rec *n = pattern_note_giveme();
		struct pattern_note_rec **ip;		/* insertion pointer */

			/* build pending note */
		*n = *pn;
		n->next = NULL;
		while (n->ofs < 0) {
			n->ofs += (curr_tempo ? curr_tempo : 1000);
		}
		n->ofs += curr_offset;

			/* insert it... */
		for (ip = &pending_notes;
		     *ip && (*ip)->ofs < n->ofs;
		     ip = &((*ip)->next))
			;
		n->next = *ip;
		*ip = n;
	}
}

static void
start_recording(void)
{
	assert_str(rec_pat != NULL, "attempted to start recording with no rec_pat.\n");

	record_going = 1;
	gettimeofday(&rec_start_time, NULL);
	printf("recording %d-measure pattern %d.\n", rec_pat->num_measures, rec_pat->patno);
}

static void
stop_recording(int from_delete)
{
	struct pattern_rec *p;
	struct pattern_note_rec *pn;
	int n = 0;

	if (!rec_pat) return;

	for (pn = rec_pat->notes; pn; pn = pn->next) {
		n++;
	}

	p = rec_pat;			/* save it for potential deletion */
	rec_pat = NULL;
	record_going = 0;
	rec_pat_note_last = NULL;

	if (n) {
		printf("done recording (%d notes).\n", n);
	} else {
		printf("done recording and no notes -- deleting!\n");
		if (!from_delete)	/* no re-entrance */
			delete_pattern(p);
	}

	determine_tempo = 0;
}

void
pattern_step(void)
{
	static int pattern_counter = 0;

	if (!playback_going) {
		return;
	}

		/* process this only every 44 calls -- i.e., divide 44.1kHz
		 * by 444, this is called about 1kHz, then */
	if (pattern_counter == 0) {
		pattern_counter = 43;
	} else {
		pattern_counter--;
		return;
	}

	if ((curr_tempo && ((curr_offset % curr_tempo) == 0)) ||
	    !curr_offset) {
		/* beginning of a new measure, add new events */
		int curr_measure = (curr_tempo ? (curr_offset/curr_tempo) : 0);
		struct pattern_rec *p;

		if (rec_pat && !record_going &&
		    ((curr_measure % rec_pat->num_measures) == 0)) {
			/* a recording was pending, just waiting for the next
			 * measure to start, so start it now. */
			start_recording();
		}

		if (playback_trace) {
			printf("beginning measure %d\n", curr_measure);
		}

		for (p = patterns; p; p = p->next) {
			if (p->muted) continue;
			if (curr_measure % p->num_measures) continue;
			pattern_to_pending(p);	/* copy to pendding queue */
		}
	}

	if (playback_trace && playback_trace_beats) {
		int i;
		int rel_ofs = curr_offset % curr_tempo;
		/* most awful way possible to do this.. */
		for (i = 0; i < playback_trace_beats; i++) {
			int beat_ofs = (i*curr_tempo)/playback_trace_beats;
			if (beat_ofs > rel_ofs) {
				break;
			} else if (beat_ofs == rel_ofs) {
				printf("beat %d/%d at %d\n", i+1, playback_trace_beats, (int)curr_offset);
			}
		}
	}

	while (pending_notes && (curr_offset >= pending_notes->ofs)) {
		char buf[128];
		struct sound_rec *old = sounds;
		struct pattern_note_rec *p = pending_notes;

		sprintf(buf,"%.80s('%s',%.3f);", p->inst, midi_notetab[p->note], ((double)p->vel)/127.0);
		cmd_doit(buf);
		if (sounds && sounds != old) {
			sounds->note_up = 44*p->duration;
		}
		if (playback_trace) {
			printf("pattern %d @%d :  %s\n", p->patno, (int)p->ofs, buf);
		}

			/* remove it from the queue.. */
		pending_notes = pending_notes->next;
		pattern_note_takeit(p);
	}

	curr_offset++;
}



void
pattern_note(unsigned char note, unsigned char attack)
{
	struct timeval tv;
	int ofs;

	gettimeofday(&tv, NULL);
	ofs = (tv.tv_sec-rec_start_time.tv_sec)*1000;
	ofs += ((1000000+tv.tv_usec)-rec_start_time.tv_usec)/1000;
	ofs -= 1000;

	if (rec_pat && record_going) {
		/* record the event in rec_pat->notes */
		if (attack) {
			/* note down, create new pattern_note */
			struct pattern_note_rec *pn;

			if (ofs > curr_tempo*rec_pat->num_measures) {
				/* done recording i guess */
				stop_recording(/*from_delete=*/0);
				return;
			}

			pn = pattern_note_giveme();
			pn->ofs = ofs;
			pn->note = note;
			pn->vel = attack;
			pn->duration = 0;	/* unknown */
			pn->inst = rec_pat->inst;
			pn->patno = rec_pat->patno;

			if (rec_pat_note_last) {
				rec_pat_note_last->next = pn;
			} else {
				rec_pat->notes = pn;
			}
			rec_pat_note_last = pn;
		} else {
			/* note up, record duration */
			struct pattern_note_rec *pn;
			for (pn = rec_pat->notes; pn; pn = pn->next) {
				if ((pn->note == note) && !pn->duration) {
					pn->duration = ofs - pn->ofs;
					break;
				}
			}
		}
	}
	if (determine_tempo) {
		/* note the tempo setting */
		curr_tempo = ofs;
		determine_tempo = 0;
		printf("tempo determined at %dms per measure\n", ofs);
	}
}


static void
pattern_tempo(int argc, char **argv)
{

	if (argc == 1) {
		/* determine the tempo from user input */
		pattern_stop();

		playback_going = 1;
		curr_offset = 0;
		curr_tempo = 0;		/* tell pattern_step() not to repeat */
		determine_tempo = 1;	/* tell pattern_note to watch */
		gettimeofday(&rec_start_time, NULL);
		printf("press a MIDI key to set the tempo.\n");
		printf("beginning playback.\n");
	} else if (argc == 2) {
		/* tempo comes from parameter */
		curr_tempo = atoi(argv[1]);
	} else {
		help("pattern tempo");
	}
}

/* reads the next integer from s, skipping over any commas between numbers */
static int
get_comma(char **s)
{
	int ret;

	if (!**s) return 0;

	ret = strtol(*s, s, 10);
	if (**s == ',') {
		(*s)++;
	}

	return ret;
}

static struct pattern_note_rec *
str_to_notes(int argc, char **argv, char *inst, int patno)
{
	struct pattern_note_rec *pn, *pn_last = NULL, *pn_first = NULL;
	int i;

	for (i = 0; i < argc; i++) {
		char *s = argv[i];

		pn = pattern_note_giveme();
		pn->next = NULL;
		pn->ofs = get_comma(&s);
		pn->note = get_comma(&s);
		pn->vel = get_comma(&s);
		pn->duration = get_comma(&s);
		pn->inst = inst;
		pn->patno = patno;

		if (pn_last) {
			pn_last->next = pn;
		} else {
			pn_first = pn;
		}
		pn_last = pn;
	}

	return pn_first;
}

static void
pattern_new(int argc, char **argv)
{
	struct pattern_rec *p;

	if (curr_pattern_explicit) {
		/* overwrite existing pattern */
		pattern_save("new");
		p = curr_pattern;
		pattern_note_list_takeit(p->notes);
		p->notes = NULL;
	} else {
		/* create new pattern */
		p = safe_malloc(sizeof *p);
		memset(p, 0, sizeof *p);
		p->cmd = "first";
		if (patterns_last) {
			p->patno = patterns_last->patno+1;
			patterns_last->next = p;
		} else {
			p->patno = 0;
			patterns = p;
		}
		patterns_last = p;
		curr_pattern = p;
	}

	if (argc >= 2) {
		p->num_measures = atoi(argv[1]);
	} else {
		p->num_measures = 1;
	}

	strncpy(p->inst, cmd_inst_getname(), 39);
	p->inst[39] = 0;

	if (argc >= 3) {
		/* the sequence is in argv */
		p->notes = str_to_notes(argc-2, argv+2, p->inst, p->patno);
	} else {
		/* record the sequence from MIDI in */
		p->notes = NULL;
		rec_pat = p;
		rec_pat_note_last = NULL;
		if (playback_going) {
			record_going = 0;
			printf("waiting for next measure break.\n");
		} else {
			start_recording();
		}
	}
}

static void
pattern_mute(void)
{
	if (!curr_pattern) {
		fprintf(stderr, "no pattern selected\n");
		return;
	}
	curr_pattern->muted = !curr_pattern->muted;
	printf("pattern %d %s.\n", curr_pattern->patno, curr_pattern->muted ? "muted" : "voiced");
}

static void
pattern_start(void)
{
	pattern_stop();

	playback_going = 1;
	curr_offset = 0;
	printf("beginning playback.\n");
}


void
pattern_stop(void)
{
	if (playback_going) {
		printf("stopping playback.\n");
	}

	playback_going = 0;
	pattern_note_list_takeit(pending_notes);
	pending_notes = NULL;

	determine_tempo = 0;

	stop_recording(/*from_delete=*/0);
}

static void
write_pattern(FILE *f, struct pattern_rec *p)
{
	struct pattern_note_rec *pn;
	char *cmd;

	if (p->stack_next) write_pattern(f, p->stack_next);

	if (p->stack_next && !strcasecmp(p->cmd, "revert")) {
		/* don't bother to save reversions that just reference
		 * previous reversions */
		return;
	}

	cmd = p->stack_next ? p->cmd : "first";

	fprintf(f, "pattern save %s %d %d %s", cmd, p->num_measures, p->muted, p->inst);
	for (pn = p->notes; pn; pn=pn->next) {
		fprintf(f, " %d,%d,%d,%d", (int)pn->ofs, (int)pn->note, (int)pn->vel, (int)pn->duration);
	}
	fprintf(f,"\n");
}

static void
pattern_write(int argc, char **argv)
{
	char *fn;
	FILE *f;

	if (argc != 2) {
		help("pattern write");
		return;
	}

	fn = extended_fn(argv[1], "pat");
	if (!(f = fopen(fn, "w"))) {
		perror(fn);
		return;
	}
	printf("writing to %s\n", fn);

	if (curr_pattern_explicit) {
		if (curr_pattern) {
			write_pattern(f, curr_pattern);
		}
	} else {
		struct pattern_rec *t;
		fprintf(f, "pattern tempo %d\n", curr_tempo);
		for (t = patterns; t; t = t->next) {
			fprintf(f, "\n");
			write_pattern(f, t);
		}
	}

	fclose(f);
}

static void
pattern_shift(int argc, char **argv)
{
	struct pattern_note_rec *pn;
	int shift;
	int start;

	if (!curr_pattern) {
		fprintf(stderr, "no pattern selected.\n");
		return;
	}
	if (!curr_pattern->notes) {
		fprintf(stderr, "empty pattern.\n");
		return;
	}

	if (argc == 2) {
		shift = atoi(argv[1]);
	} else if (argc == 1) {
		shift = -curr_pattern->notes->ofs;
	} else {
		help("pattern shift");
		return;
	}

	pattern_save("shift");

	start = curr_pattern->notes->ofs + shift;

	for (pn = curr_pattern->notes; pn; pn = pn->next) {
		pn->ofs += shift;
		if (pn->ofs < start) {
			start = pn->ofs;
		}
	}

	printf("pattern %d now begins at %dms (shift=%d)\n", curr_pattern->patno, start, shift);
}

static void
pattern_inst(int argc, char **argv)
{
	if (!curr_pattern) {
		fprintf(stderr, "no pattern selected\n");
		return;
	}
	printf("pattern %d instrument was %s.\n", curr_pattern->patno, curr_pattern->inst);
	if (argc == 1) {
		pattern_save("inst");
		strncpy(curr_pattern->inst, cmd_inst_getname(), 39);
		curr_pattern->inst[39] = 0;
	} else if (argc == 2) {
		pattern_save("inst");
		strncpy(curr_pattern->inst, argv[1], 39);
		curr_pattern->inst[39] = 0;
	} else {
		help("pattern inst");
	}
	printf("now %s.\n", curr_pattern->inst);
}

static void
pattern_measures(int argc, char **argv)
{
	if (!curr_pattern) {
		fprintf(stderr, "no pattern selected\n");
		return;
	}
	if (argc != 2) {
		help("pattern measures");
		return;
	}
	pattern_save("measures");
	printf("pattern %d was %d measures.\n", curr_pattern->patno, curr_pattern->num_measures);
	curr_pattern->num_measures = atoi(argv[1]);
	printf("now %d measures.\n", curr_pattern->num_measures);
}

static int
note_ofs(char *arg, int old_ofs)
{
	static int ofs = 0;
	double argval;
	argval = strtod((isalpha(arg[0]) ? &arg[1] : &arg[0]), NULL);
	if (arg[0] == 'o') {
		ofs = old_ofs;
	} else if (arg[0] == 'x') {
		ofs += (((double)curr_tempo)*argval) + 0.5;
	} else if (arg[0] == 'm') {
		ofs = (((double)curr_tempo)*argval) + 0.5;
	} else if (argval != 0.0 && argval != -0.0) {
		ofs += (((double)curr_tempo)/argval) + 0.5;
	}
	return ofs;
}

static void
pattern_rhythm(int argc, char **argv)
{
	struct pattern_note_rec *pn;
	int i;
	int n = 0;

	note_ofs("m0", 0);		/* reset to beginning */

	if (!curr_pattern) {
		fprintf(stderr, "no pattern selected\n");
		return;
	}
	if (argc < 2) {
		help("pattern rhythm");
		return;
	}

	pattern_save("rhythm");

	printf("old:\n");
	pattern_note_list(curr_pattern->notes, curr_pattern->num_measures);

	i = 1;
	for (pn = curr_pattern->notes; pn; pn = pn->next) {
		pn->ofs = note_ofs(argv[i++], pn->ofs);
		if (i >= argc) {
			if (argc == 2) {
				/* just a single parameter, use it as
				 * spacing for all notes */
				i = 1;
			} else {
				/* this is the last parameter, so add
				 * the first parameter too */
				pn->ofs = note_ofs(argv[1], pn->ofs);
				i = 2;
			}
		}
		n++;
	}

	printf("%d notes of pattern %d assigned:\n", n, curr_pattern->patno);
	pattern_note_list(curr_pattern->notes, curr_pattern->num_measures);
}

static void
pattern_velocity(int argc, char **argv)
{
	struct pattern_note_rec *pn;
	int i = 1;
	int n = 0;

	if (!curr_pattern) {
		fprintf(stderr, "no pattern selected\n");
		return;
	}
	if (argc < 2) {
		help("pattern velocity");
		return;
	}

	pattern_save("velocity");

	for (pn = curr_pattern->notes; pn; pn = pn->next) {
		pn->vel = strtod(argv[i++], NULL)*127.0;
		if (i >= argc) {
			i = 1;
		}
		n++;
	}

	printf("assigned new velocities to %d notes in pattern %d.\n", n, curr_pattern->patno);
}

static int
stack_depth(struct pattern_rec *p)
{
	int i = 0;
	for (; p; p = p->stack_next) {
		i++;
	}
	return i;
}

static void
pattern_revert(int argc, char **argv)
{
	struct pattern_rec *p, *new_p;
	int i;
	int n;

	if (!curr_pattern) {
		fprintf(stderr, "no pattern selected\n");
		return;
	}

	n = stack_depth(curr_pattern);

	if (argc == 1) {
		i = 1;		/* pop one state */
	} else if (argc == 2) {
		i = (n-1)-atoi(argv[1]);
	} else {
		help("pattern revert");
		return;
	}

	if (i <= 0) {
		printf("tried to negative pop, doing nothing\n");
		return;
	}
	if (i >= n) {
		printf("tried to pop past end of stack, reverting to original\n");
		i = n-1;
	}

	printf("popping %d modifications off of pattern %d to get to stack level %d\n", i, curr_pattern->patno, (n-1)-i);

	for (p = curr_pattern; p; p = p->stack_next) {
		if (!i--)
			break;
	}

	assert_str(p != NULL, "really walked past end of stack, oops");

	new_p = safe_malloc(sizeof *new_p);
	*new_p = *curr_pattern;
	pattern_copy(curr_pattern, p);
	curr_pattern->next = new_p->next;
	curr_pattern->stack_next = new_p;
	new_p->next = NULL;
	curr_pattern->cmd = "revert";
}

static void
pattern_save_cmd(int argc, char **argv)
{
	struct pattern_rec *p;
	int first = (argc >= 2 && !strcasecmp(argv[1], "first"));

	if (!curr_pattern && !(argc > 2 && !strcasecmp(argv[1], "first"))) {
		fprintf(stderr, "no pattern selected.\n");
		return;
	}
	if (argc < 5) {
		help("pattern save");
		return;
	}

	p = safe_malloc(sizeof *p);
	if (first) {
		memset(p, 0, sizeof *p);
	} else {
		*p = *curr_pattern;
		p->next = NULL;
		curr_pattern->stack_next = p;
		p = curr_pattern;
	}
	p->cmd = safe_strdup(argv[1]);		/* leaks */
	p->num_measures = atoi(argv[2]);
	p->muted = atoi(argv[3]);
	strncpy(p->inst, argv[4], 39);
	p->inst[39] = 0;
	p->notes = str_to_notes(argc-5, argv+5, p->inst, p->patno);
	if (first) {
		p->next = NULL;
		p->stack_next = NULL;
		if (patterns_last) {
			p->patno = patterns_last->patno+1;
			patterns_last->next = p;
		} else {
			p->patno = 0;
			patterns = p;
		}
		patterns_last = p;
		curr_pattern = p;
	}

}

static void
pattern_list(int argc, char **argv)
{
	int max_measures = 0;
	struct pattern_rec *p;

	if (curr_pattern_explicit) {
		int i = stack_depth(curr_pattern)-1;
		for (p = curr_pattern; p; p = p->stack_next) {
			printf("pattern %d stack %d: %s (%d measures, \"%s\" instrument)%s%s\n", p->patno, i--, p->cmd, p->num_measures, p->inst, p->muted ? " MUTED" : "", (p == curr_pattern) ? " CURRENT" : "");
			pattern_note_list(p->notes, p->num_measures);
		}
	} else {
		for (p = patterns; p; p = p->next) {
			if (p->num_measures > max_measures)
				max_measures = p->num_measures;
		}

		printf("tempo: %d\n", curr_tempo);
		printf("pending:\n");
		pattern_note_list(pending_notes, max_measures);

		for (p = patterns; p; p = p->next) {
			printf("pattern %d stack %d: %s (%d measures, \"%s\" instrument)%s%s\n", p->patno, stack_depth(p)-1, p->cmd, p->num_measures, p->inst, p->muted ? " MUTED" : "", (p == curr_pattern) ? " CURRENT" : "");
			pattern_note_list(p->notes, max_measures);
		}
	}
}


static void
pattern_delete(int argc, char **argv)
{
	if (argc != 1) {
		help("pattern delete");
		return;
	}
	if (!curr_pattern) {
		fprintf(stderr, "no pattern selected\n");
		return;
	}

	delete_pattern(curr_pattern);
}

static char *
pattern_edit_write(void)
{
	struct pattern_note_rec *n;
	int ofs = 0;
	char *fn = "buh.tmp";
	FILE *f = fopen(fn,"w");
	if (!f) {
		perror(fn);
		return NULL;
	}
	for (n = curr_pattern->notes; n; n = n->next) {
		if (n->ofs > ofs) {
			fprintf(f,"S %d\n",n->ofs-ofs);
			ofs = n->ofs;
		}
		fprintf(f,"N %s %d %d # N note vel dur. ofs=%d\n",midi_notetab[n->note], n->vel, n->duration, (int)n->ofs);
	}
	fclose(f);

	return fn;
}

static void
pattern_edit_read(char *fn)
{
	int ofs = 0;
	FILE *f = fopen(fn,"r");
	struct pattern_note_rec *n, *n_last = NULL;
	char buf[1024];
	char *p;
	if (!f) {
		perror(fn);
		return;
	}
	pattern_note_list_takeit(curr_pattern->notes);
	curr_pattern->notes = 0;
	while (fgets(buf, 1024, f)) {
		p = buf;
		if (*p == 'S') {
			p ++;
			while (isspace(*p)) p++;
			ofs += atoi(p);
		} else if (*p == 'N') {
			int i;
			n = pattern_note_giveme();
			n->inst = curr_pattern->inst;
			n->patno = curr_pattern->patno;
			p++;
			while (isspace(*p)) p++;
			for (i = 0; i < sizeof midi_notetab/sizeof midi_notetab[0]; i++) {
				if (!memcmp(p,midi_notetab[i],3)) {
					break;
				}
			}
			n->ofs = ofs;
			n->note = i;
			p+=3;
			while (isspace(*p)) p++;
			n->vel = strtoul(p, &p, 10);
			while (isspace(*p)) p++;
			n->duration = strtoul(p, &p, 10);
			n->next = NULL;

			if (n_last) {
				n_last->next = n;
			} else {
				curr_pattern->notes = n;
			}
			n_last = n;
		}
	}


}

static void
pattern_edit(int argc, char **argv)
{
	char *fn;
	if (argc != 1) {
		help("pattern edit");
		return;
	}
	if (!curr_pattern) {
		fprintf(stderr, "no pattern selected\n");
		return;
	}
	fn = pattern_edit_write();
	{
		char buf[1024];
		sprintf(buf,"nvi %s\n",fn);
		system(buf);
	}
	pattern_save("edit");
	pattern_edit_read(fn);
}


/* displays the events in the given pattern note chain in a nice graphical
 * format, with enough space to fit a certain number of measures */
/* XXX - make it so if measures > num_measures, it shows the looping... */
static void
pattern_note_list(struct pattern_note_rec *pn, int measures)
{
	int i;
	struct pattern_note_rec *pns[80];
	struct pattern_note_rec conflict;
	int m;

	if (!measures) return;

	memset(pns, 0, sizeof pns);

	/* XXX - is there a way to get the number of columns from readline? */
	for (; pn; pn = pn->next) {
		int col = (80.0 * pn->ofs) / ((double)(curr_tempo * measures));
		if (col >= 80) col = 79;
		if (col < 0) col = 0;
		if (pns[col]) {
			pns[col] = &conflict;
		} else {
			pns[col] = pn;
		}
	}

	m = -1;
	for (i = 0; i < 80; i++) {
		char c = '-';
		int curr_m = ((double)i*measures)/80.0;
		if (curr_m > m) {
			m = curr_m;
			if (curr_m < 10) {
				c = curr_m+'0';
			} else if (curr_m < 36) {
				c = curr_m-10+'a';
			} else {
				c = '*';
			}
		}
		printf("%c",c);
	}
	printf("\n");
	for (i = 0; i < 80; i++) {
		char c = ' ';
		if (pns[i] == &conflict) {
			c = 'X';
		} else if (pns[i]) {
			/* XXX - it should show the note value */
			c = 'n';
		}
		printf("%c",c);
	}
	/* XXX - second row for note octave? */
	printf("\n");
}


static void
pattern_copy(struct pattern_rec *dst, struct pattern_rec *src)
{
	struct pattern_note_rec *pn_start, *pn_last, *pn;

	*dst = *src;
	dst->next = NULL;
	dst->stack_next = NULL;

	pn_start = NULL;
	pn_last = NULL;
	for (pn = curr_pattern->notes; pn; pn = pn->next) {
		struct pattern_note_rec *pn_new = pattern_note_giveme();
		*pn_new = *pn;
		pn_new->next = NULL;
		if (pn_last) {
			pn_last->next = pn_new;
		} else {
			pn_start = pn_new;
		}
		pn_last = pn_new;
	}
	dst->notes = pn_start;
}

/* for memory management purposes, the argument should probably be just
 * a string literal, or it will be leaked */
static void
pattern_save(char *cmd)
{
	struct pattern_rec *p;

	assert_str(curr_pattern != NULL, "trying to save NULL pattern");

	p = safe_malloc(sizeof *p);
	pattern_copy(p, curr_pattern);
	p->stack_next = curr_pattern->stack_next;
	p->next = NULL;
	curr_pattern->stack_next = p;
	curr_pattern->cmd = cmd;
}

static void
delete_pattern(struct pattern_rec *pat)
{
	struct pattern_rec *last = NULL;
	struct pattern_rec *p;
	struct pattern_rec *p_next;

	for (p = patterns; p; p = p->next) {
		if (p == pat) break;		/* gotcha */
		last = p;
	}

	assert_str(p == pat, "delete_pattern() failed!!");

	if (last) {
		last->next = pat->next;
	} else {
		patterns = pat->next;
	}

	if (patterns_last == pat) {
		patterns_last = last;
	}
	if (curr_pattern == pat) {
		curr_pattern = NULL;
	}
	if (rec_pat == pat) {
		stop_recording(/*from_delete=*/1);
	}

	for (p = pat; p; p = p_next) {
		p_next = p->stack_next;
		pattern_note_list_takeit(p->notes);
		free(p);
	}
}

static void
pattern_trace(int argc, char **argv)
{
	if (argc == 1) {
		/* just toggle it */
		playback_trace = !playback_trace;
		playback_trace_beats = 0;
	} else if (argc == 2) {
		/* set the number of beats per measure, turn it on */
		playback_trace_beats = atoi(argv[1]);
		playback_trace = 1;
	} else {
		help("pattern trace");
	}
	printf("pattern tracing %s.\n", (playback_trace ? "enabled" : "disabled"));
}

static void
pattern_reset(void)
{
	struct pattern_rec *p, *p_next;

	pattern_stop();
	stop_recording(/*from_delete=*/0);

	for (p = patterns; p; p = p_next) {
		p_next = patterns->next;
		delete_pattern(p);
	}

}

/* XXX - make 'pattern start <patno>' */
/* XXX - make sure all curr_pattern refs check for null */
/* XXX - make a higher default tempo value for easier boot strapping?, or make an infinite record function? */
/* XXX - pattern stack */
/* XXX - pattern duration (uses note_ofs(), probably) */
/* XXX - pattern note changer somehow, maybe an inverter?, or a minorer? */
/* XXX - automatically delete empty patterns */
