/* Copyright 2010 Greg Alexander.  See the file COPYING for terms */
/* $Id: tracker.c,v 1.20 2010-02-08 03:17:28 greg Exp $ */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include "misc.h"
#include "tracker.h"
#include "snd.h"
#include "lang.h"

#define MAX_LINE_LEN 10240


/* records pending note up events */
struct noteup_rec {
	struct noteup_rec *next;
	int sound_id;			/* id # of sound_rec */
	int ticks_remaining;		/* how long until we do the note up */
};


/* just a linklist structure for extrack event strings */
struct events_list {
	struct events_list *next;
	int noteup;			/* -1 = no noteup definition */
	char evt[0];
};

/* each entry in a dictionary is either a track
 * (references another dictionary) or a command to execute immediately: */
struct trk_dict_ent {
	enum {
		tde_undef = 0,
		tde_extrack,
		tde_track,
		tde_command
	} kind;
	union {
		struct {
			unsigned ticks_per;
			struct events_list *events;
		} extrack;
		struct {
			char *dict;
			unsigned ticks_per;
			char *events;
			char *args[10];
		} track;
		struct {
			int noteup;	/* -1 = no noteup specified */
			char *command;	/* the command to pass to cmd_doit() */
		} command;
	} u;
};

/* a list of dictionaries */
struct trk_dict {
	struct trk_dict *next;
	char *name;
	struct trk_dict_ent ents[52];	/* mapped a-zA-Z */
	char *def_args[10];		/* the default values for args 0-9 */
};
#define LETTER_TO_INDEX(x) (islower(x) ? ((x) - 'a') : (((x) - 'A') + 26))

/* this is just a raw chain of events (essentially a sequence of
 * tde_command trk_dict_ents) for playback, the flattened output from
 * treeing... */
struct events_rec {
	struct events_rec *next;
	unsigned tick_ofs;
	int noteup;		/* -1 = no noteup specified */
	char command[0];	/* string to execute for the event */
};

static struct trk_dict *dicts = NULL;
static struct noteup_rec *noteups = NULL;
static struct events_rec *events = NULL;

static struct events_rec *curr_event = NULL;
static unsigned curr_tick = 0;
static unsigned curr_per_tick = 0;
static unsigned curr_countdown = 0;


/* enjoy static functions */
static void set_curr_event(void);
static int load_file(char *fn);
static struct trk_dict_ent *process_line(char *line, int line_no);
static int process_extrk_line(char *line, int line_no, struct trk_dict_ent *ent);
static void add_noteup(int id, int sustain);
static void free_dicts(void);
static struct trk_dict_ent *new_dict_ent(char *name);
static struct trk_dict *dict_lookup(char *name, int define);
static void flatten_ent(struct trk_dict *dict, struct trk_dict_ent *ent, unsigned tick_ofs, char **args, int recur);

/* tell the tracker to start playing a new song, in the file named by
 * fn.  if loop is set, it will loop indefinitely (for experimentation
 * with loops).  Skip is the number of ticks at the beginning of the
 * song to skip. */
void
tracker_play(char *fn, int loop, int skip)
{
	/* make sure previous track is unloaded.. */
	tracker_stop();

	/* clear out whitespace */
	while(isspace(*fn)) fn++;
	fn[strcspn(fn," \t\r\n")] = 0;

	fn = extended_fn(fn, "trk");

	if (!load_file(fn))
		return;
	curr_tick = 0;
	set_curr_event();
	tracker_rate(60);
	curr_countdown = 0;

	/* execute the first timestep, so that definitions get loaded
	 * if we're using skip */
	tracker_timestep();

	curr_tick = skip;
	set_curr_event();
}


/* initializes curr_event to point into the appropriate part of the
 * events list, given the value of curr_tick */
static void
set_curr_event(void)
{
	for (curr_event = events; curr_event; curr_event = curr_event->next) {
		if (curr_event->tick_ofs >= curr_tick)
			return;
	}
}


/* reads in a tracker file, returning false (0) on failure */
static int
load_file(char *fn)
{
	FILE *f;
	char buf[MAX_LINE_LEN];
	int line_no = 0;
	struct trk_dict *main_dict;
	int i;

	if (!(f = open_search(fn,"r"))) {
		perror(fn);
		return 0;
	}

	free_dicts();

	while (fgets(buf,MAX_LINE_LEN,f)) {
		struct trk_dict_ent *extrk;

		extrk = process_line(buf, ++line_no);
		if (extrk) {
			while (fgets(buf, MAX_LINE_LEN, f)) {
				if (!process_extrk_line(buf, ++line_no, extrk))
					break;
			}
		}
	}

	fclose(f);

	main_dict = dict_lookup("main", /*define=*/0);
	if (!main_dict) {
		fprintf(stderr,"no main dictionary... playing nothing.\n");
		free_dicts();
		return 0;
	}

	for (i = 0; i < 52; i++) {
		flatten_ent(main_dict, &(main_dict->ents[i]),0,main_dict->def_args, /*recur=*/0);
	}

	return 1;
}


/* a default dictionary parameter line of the form "dict_0: argument" */
static void
process_defarg_line(char *line)
{
	int argno;
	struct trk_dict *dict;
	char *name = line;

	line = strchr(line, '_');
	assert_str(line != NULL, "no underscore in defarg line?");

	argno = line[1]-'0';
	assert_str(argno >= 0 && argno <= 9, "bad defarg number in defarg line?");

	*line = 0;
	line += 3;
	line += strspn(line, " \t");
	dict = dict_lookup(name, /*define=*/1);
	if (dict->def_args[argno]) {
		fprintf(stderr,"apparent redefinition of %s_%d\n",name,argno);
		free(dict->def_args[argno]);
		dict->def_args[argno] = NULL;
	}
	if (*line) {
		dict->def_args[argno] = safe_strdup(line);
	}
}

/* a command definition line, where name is like "dict_a" and cmd is the
 * command */
static void
process_cmd_line(char *name, char *cmd)
{
	struct trk_dict_ent *ent;
	char *at;

	ent = new_dict_ent(name);
	ent->kind = tde_command;
	ent->u.command.noteup = -1;

	cmd += strspn(cmd, " \t");
	if ((at = strchr(cmd,'@'))) {
		*at++ = 0;
		at += strspn(at," \t");
		if (isdigit(*at)) {
			ent->u.command.noteup = atoi(at);
		}
	}
	ent->u.command.command = safe_strdup(cmd);
}

/* processes the first line of an extended track definition, name should be
 * like "dict_a", and line should be like "8 {"
 */
static struct trk_dict_ent *
process_extrk_start(char *name, char *line, int line_no)
{
	struct trk_dict_ent *ent;
	int ticks_per;

	ticks_per = strtoul(line, &line, 10);
	line += strspn(line, " \t");
	if (*line != '{') /* } */ {
		fprintf(stderr,"line %d: expecting '{', perhaps a dictionary name was omitted?\n",line_no); /* } */
		return NULL;
	}
	line++;
	line += strspn(line, " \t\r\n");
	if (*line) {
		fprintf(stderr,"line %d: trailing garbage ignored\n",line_no);
	}

	ent = new_dict_ent(name);
	ent->kind = tde_extrack;
	ent->u.extrack.ticks_per = ticks_per;
	ent->u.extrack.events = NULL;
	ent->u.track.events = safe_strdup(line);
	return ent;
}

/* line should be "(arg1, arg2, ...)...", the arguments in the parantheses
 * are parsed into the args[] array (which must be a pointer to an array
 * of ten pointers) and a pointer to immediately after the closing paren
 * is returned */
static char *
process_args(char *line, char **args, int line_no)
{
	static char *local = NULL;
	static int local_len = 0;
	int argno = 0;
	char *arg;
	char *end;

		/* copy it to a local buffer because we're going to modify it */
	if (strlen(line) >= local_len) {
		free(local);
		local_len = strlen(line) + 100;
		local = safe_malloc(local_len+1);
	}
	strcpy(local, line);
	line = local;

	assert_str(*line == '(', "process_args with non-args?");
	line++;
	end = strchr(line, ')');
	if (!end) {
		fprintf(stderr,"line %d: no closing paren\n", line_no);
		return line;
	}
	*end++ = 0;

	arg = line;
	while (*arg) {
		char *arg_next = arg + strcspn(arg, ",");

		if (*arg_next) *arg_next++ = 0;

		if (*arg) {
			if (argno >= 10) {
				fprintf(stderr,"line %d: too many args (max 10)\n", line_no);
				return end;
			}
			args[argno] = safe_strdup(arg);
		}

		argno++;
		arg = arg_next;
	}

	return end;
}

/* process a normal sequence definition line, where name is like "dict1_a"
 * and line is like "dict2 8 [abc...xyz]"
 */
static void
process_trk_line(char *name, char *line, int line_no)
{
	char *fromdict;
	unsigned ticks_per;
	char *evt_end;
	struct trk_dict_ent *ent;
	char *args[10];

	memset(args,0,sizeof args);

	fromdict = line;
	while (isalnum(*line)) line++;

		/* arguments for the from dictionary... */
	if (*line == '(') {
		char *args_start = line;
		line = process_args(line, args, line_no);
		*args_start = 0;	/* terminate the dictionary name */
	}

	if (fromdict == line || !isspace(*line)) {
		fprintf(stderr,"line %d: expected a from dictionary name, is this not a track?\n",line_no);
		return;
	}
	*line++ = 0;
	line += strspn(line," \t");

	if (!isdigit(*line)) {
		fprintf(stderr,"line %d: expected a tick count, is this not a track?\n",line_no);
		return;
	}
	ticks_per = strtoul(line, &line, 10);
	line += strspn(line, " \t");
	if (*line != '[' || !(evt_end = strchr(line,']'))) {
		fprintf(stderr,"line %d: expected [events], is this not a track?\n",line_no);
		return;
	}
	line++;
	*evt_end = 0;

	ent = new_dict_ent(name);
	ent->kind = tde_track;
	ent->u.track.dict = safe_strdup(fromdict);
	ent->u.track.ticks_per = ticks_per;
	ent->u.track.events = safe_strdup(line);
	memcpy(ent->u.track.args, args, sizeof args);
}

/* returns the defined dictionary entry if it is an tde_extrack entry
 * and still needs more lines to be scanned... */
static struct trk_dict_ent *
process_line(char *line, int line_no)
{
	char *name;
	struct trk_dict_ent *retval = NULL;

	line += strspn(line," \t");
	line[strcspn(line,"\r\n")] = 0;
	if (!*line || *line == '#') {
		return NULL;
	}

	name = line;
	while (isalnum(*line)) line++;
	if (name == line || *line != '_') {
		fprintf(stderr,"line %d: should begin with dict_X\n",line_no);
		return NULL;
	}
	line++;

	if (isdigit(line[0]) && line[1] == ':') {
		/* this is a default argument definition, like main_0: 1.0 */
		/* handle it here then return immediately so we don't need to
		 * worry about this special case definition */
		process_defarg_line(name);
		return NULL;
	}

	if (!isalpha(*line)) {
		fprintf(stderr,"line %d: should begin with dict_X\n",line_no);
		return NULL;
	}
	line++;
	if (*line == ':') {
		*line++ = 0;
		process_cmd_line(name, line);
	} else {
		if (!isspace(*line)) {
			fprintf(stderr,"line %d: should begin with dict_X[:]\n",line_no);
			return NULL;
		}
		*line++ = 0;
		line += strspn(line, " \t");

		if (isdigit(*line)) {
			/* most likely of the form "dict_x 8 {"     } */
			retval = process_extrk_start(name, line, line_no);
		} else {
			/* of the form "dict_x dict 8 [abc...xyz]" */
			process_trk_line(name, line, line_no);
		}
	}
	return retval;
}

static int
process_extrk_line(char *line, int line_no, struct trk_dict_ent *ent)
{
	static struct events_list **event_last = NULL;
	struct events_list *e;
	int noteup;
	char *at;

	line += strspn(line," \t");
	line[strcspn(line,"\r\n")] = 0;

/* { */
	if (*line == '}') {
		line++;
		line += strspn(line," \t");
		if (*line) {
			fprintf(stderr,"line %d: ignoring trailing garbage\n",line_no);
		}
		event_last = NULL;
		return 0;
	}

	noteup = -1;
	if ((at = strchr(line,'@'))) {
		*at++ = 0;
		at += strspn(at," \t");
		if (isdigit(*at)) {
			noteup = atoi(at);
		}
	} else {
		noteup = -1;
	}

	e = safe_malloc(strlen(line) + 1 + sizeof *e);
	e->next = NULL;
	e->noteup = noteup;
	strcpy(e->evt, line);

	if (!event_last) {
		event_last = &ent->u.extrack.events;
	}

	*event_last = e;
	event_last = &e->next;

	return 1;
}


void
tracker_timestep()
{
	struct noteup_rec **np;

	if (!curr_countdown) {
		while (curr_event && curr_event->tick_ofs <= curr_tick) {
			char *cmd = curr_event->command;
			int noteup = curr_event->noteup;
			struct sound_rec *old = sounds;
			if (cmd) {
				cmd_doit(cmd);
				if (noteup >= 0 && sounds && sounds != old) {
					add_noteup(sounds->id, noteup);
				}
			}
			curr_event = curr_event->next;
		}

		for (np = &noteups; *np; ) {
			struct noteup_rec *n = *np;
			if (n->ticks_remaining == 0) {
				*np = n->next;
				snd_noteup(n->sound_id);
				free(n);
			} else {
				n->ticks_remaining--;
				np = &(n->next);
			}
		}

		curr_tick++;
		curr_countdown = curr_per_tick;

		if (curr_tick == 1 && curr_event) {
			precompile();
		}
	}
	if (curr_countdown) curr_countdown--;
}

void
tracker_cleanup(void)
{
	tracker_stop();
}

void
tracker_rate(unsigned rate)
{
	curr_per_tick = (44100*60 + (rate / 2)) / rate;
	if (curr_countdown >= curr_per_tick) curr_countdown = curr_per_tick;
}

static void
add_noteup(int id, int sustain)
{
	struct noteup_rec *n = safe_malloc(sizeof *n);
	n->next = noteups;
	n->sound_id = id;
	n->ticks_remaining = sustain;
	noteups = n;
}


/* free memory pointed to from a single trk_dict_ent record */
static void
free_dict_ent(struct trk_dict_ent *ent)
{
	int i;
	struct events_list *e, *e_n;

	switch (ent->kind) {
		case tde_undef: break;	/* nothing */
		case tde_extrack:
			for (e = ent->u.extrack.events; e; e = e_n) {
				e_n = e->next;
				free(e);
			}
			break;
		case tde_track:
			free(ent->u.track.dict);
			free(ent->u.track.events);
			for (i = 0; i < 10; i++) {
				free(ent->u.track.args[i]);
			}
			break;
		case tde_command:
			free(ent->u.command.command);
			break;
	}
}


/* clears out all the stuff hanging off of dicts */
/* also gets rid of events, because it's all based off of dicts */
static void
free_dicts(void)
{
	struct events_rec *e, *e_n;
	struct trk_dict *d, *d_n;
	int i;

	for (e = events; e; e = e_n) {
		e_n = e->next;
		free(e);
	}
	events = NULL;
	curr_event = NULL;

	for (d = dicts; d; d = d_n) {
		d_n = d->next;
		for (i = 0; i < 52; i++) {
			free_dict_ent(&d->ents[i]);
		}
		for (i = 0; i < 10; i++) {
			free(d->def_args[i]);
		}
		free(d->name);
		free(d);
	}
	dicts = NULL;
}


/* returns a pointer to a dictionary letter entry to be filled in */
/* name is of the form like "dict_a" */
static struct trk_dict_ent *
new_dict_ent(char *name)
{
	char *s = name;
	struct trk_dict *dict;
	int wordno;

	while (*s && *s != '_') s++;
	if (!*s || !isalpha(s[1])) return NULL;
	*s++ = 0;
	wordno = LETTER_TO_INDEX(*s);

	dict = dict_lookup(name, /*define=*/1);

	if (dict->ents[wordno].kind != tde_undef) {
		fprintf(stderr,"apparent redefinition of %s_%c\n",name,*s);
	}

	return &(dict->ents[wordno]);
}

/* Looks up a dictionary by name.  On failure, returns NULL unless define
 * is set, in which case it defines an empty dictionary entry */
static struct trk_dict *
dict_lookup(char *name, int define)
{
	struct trk_dict *dict = NULL;
	for (dict = dicts; dict; dict = dict->next) {
		if (!strcmp(dict->name, name))
			return dict;
	}
	if (define) {
		dict = safe_malloc(sizeof *dict);
		memset(dict,0,sizeof *dict);
		dict->next = dicts;
		dict->name = safe_strdup(name);
		dicts = dict;
	}
	return dict;
}

/* adds an event to the global event list, to occur at tick_ofs, with
 * a note up event after noteup ticks and with the specified command.
 * the caller is responsible for any memory management on command. */
static void
add_event(unsigned tick_ofs, int noteup, char *command)
{
	struct events_rec **p, *n;

	n = safe_malloc(strlen(command) + 1 + sizeof *n);
	n->tick_ofs = tick_ofs;
	n->noteup = noteup;
	strcpy(n->command, command);
	for (p = &events; *p; p = &((*p)->next)) {
		if ((*p)->tick_ofs >= tick_ofs)
			break;
	}
	n->next = *p;
	*p = n;
}

/* process a line that begins with like "dict_a" and return a pointer to
 * immediately after the 'a'.  Place a pointer to a static buffer holding
 * "dict" in *dictname, and the 'a' in *ent. */
static char *
process_dictent(char *line, char **dictname, char *ent)
{
	static char *dict = NULL;
	static int dict_sz = 0;
	int dict_len;
	char *s;

		/* defaults */
	*dictname = "";
	*ent = 'a';

	line += strspn(line, " \t");
	s = line;

	while (isalnum(*s)) s++;
	if ((s == line) || (s[0] != '_') || !isalpha(s[1])) {
		fprintf(stderr,"%s is not a proper dictionary entry\n",line);
		return s;
	}
	dict_len = s-line;
	if (dict_len+1 >= dict_sz) {
		dict_sz = dict_len + 128;
		free(dict);
		dict = safe_malloc(dict_sz);
	}
	memcpy(dict, line, dict_len);
	dict[dict_len] = 0;

	*dictname = dict;
	*ent = s[1];

	return &s[2];
}

/* call is either like "dict_a" or like "dict_a(args)".  this uses
 * flatten_ent() to add entries to events appropriate to that dictionary
 * invocation. */
static void
call_dictent(char *call, unsigned tick_ofs, int recur)
{
	char *dictname;
	char ent;
	char **args;
	char *xargs[10];
	int wordno;
	struct trk_dict *dict;

	call = process_dictent(call, &dictname, &ent);
	call += strspn(call," \t");

	dict = dict_lookup(dictname, /*define=*/0);
	wordno = LETTER_TO_INDEX(ent);

	if (!dict || (dict->ents[wordno].kind == tde_undef)) {
		fprintf(stderr,"%s_%c undefined, skipping!\n", dictname, ent);
		return;
	}

	args = dict->def_args;

	if (*call == '(') {
		int i;

		call = process_args(call, xargs, -1);

		for (i = 0; i < 10; i++) {
			if (!xargs[i] && args && args[i]) {
		/* xargs[] is all memory we are responsible for, so we have to
		 * strdup this just to accomodate the free() in its future */
				xargs[i] = safe_strdup(args[i]);
			}
		}
		args = xargs;
	}

	call += strspn(call," \t");
	if (*call) {
		fprintf(stderr,"ignoring trailing garbage\n");
	}

	flatten_ent(dict, &dict->ents[wordno], tick_ofs, args, recur+1);

		/* if xargs was used, it needs to be freed */
	if (args == xargs) {
		int i;
		for (i = 0; i < 10; i++) {
			free(xargs[i]);
		}
	}
}

/* returns a pointer to a static buffer that holds the expanded version
 * of str, with args applied to it
 */
static char *
arg_replace(char *str, char **args)
{
	/* may I present a lazy man's variable length string: */
#define EXPAND_BUF(x) if ((x) >= buf_sz) { char *__t; int __nl = (x) + 128; __t = safe_malloc(__nl); memcpy(__t, buf, buf_sz); buf_sz = __nl; free(buf); buf = __t; }
	static char *buf = NULL;
	static int buf_sz = 0;
	int buf_len = 0;

	assert_str(str != NULL, "arg_replace passed null pointer?");

	while (*str) {
		if (str[0] == '$' && isdigit(str[1])) {
			/* argument substitution */
			int argno = str[1]-'0';
			str += 2;
			if (args && args[argno]) {
				int alen = strlen(args[argno]);
				EXPAND_BUF(buf_len + alen)
				memcpy(buf+buf_len, args[argno], alen);
				buf_len += alen;
			}
		} else {
			EXPAND_BUF(buf_len)
			buf[buf_len] = *str;
			buf_len++;
			str++;
		}
	}

		/* null-terminate */
	EXPAND_BUF(buf_len)
	buf[buf_len] = 0;

	return buf;
}

/* Evaluate 'ent' into the events list starting at tick_ofs, with
 * dictionary parameters in args[].  Note that the caller is responsible
 * for any memory freeing that may be necessary on members of args[].
 * dict is the dictionary that ent came from, used to determine if
 * a dictionary is calling itself and therefore should inherit arguments. */
static void
flatten_ent(struct trk_dict *dict, struct trk_dict_ent *ent, unsigned tick_ofs, char **args, int recur)
{
	if (recur > MAX_RECUR) {
		fprintf(stderr, "recursion overflow in tracker!\n");
		return;
	}
	switch (ent->kind) {
		case tde_undef: break;	/* nothing */
		case tde_extrack: {
			struct events_list *e;
			for (e = ent->u.extrack.events; e; e = e->next) {
				if (e->evt[0] == ':') {
					/* :dict_a(args) */
					call_dictent(&e->evt[1], tick_ofs, recur);
				} else if (e->evt[0]) {
					/* command */
					add_event(tick_ofs,
						  e->noteup,
						  arg_replace(e->evt, args));
				}
				tick_ofs += ent->u.extrack.ticks_per;
			}
		}	break;
		case tde_track: {
			struct trk_dict *sdict = dict_lookup(ent->u.track.dict, /*define=*/0);
			char *s = ent->u.track.events;
			char *xargs[10];

			if (!sdict) {
				fprintf(stderr,"dictionary \"%s\" not found, skipping track!\n",ent->u.track.dict);
				return;
			}
				
			if (dict == sdict) {
				/* the default arguments should be the ones
				 * we were passed, let them fall through */
			} else {
				/* the default arguments should be the ones
				 * of the destination dictionary */
				args = sdict->def_args;
			}

				/* merge the specific arguments with the
				 * default arguments in xargs[], notably
				 * not allocating any new memory */
			if (ent->u.track.args) {
				int i;
				for (i = 0; i < 10; i++) {
					if (ent->u.track.args[i]) {
						xargs[i] = ent->u.track.args[i];
					} else if (args[i]) {
						xargs[i] = args[i];
					} else {
						xargs[i] = NULL;
					}
				}
				args = xargs;
			}

			while (*s) {
				struct trk_dict_ent *ent2;

				if (isalpha(*s)) {
					ent2 = &(sdict->ents[LETTER_TO_INDEX(*s)]);
					if (ent2->kind == tde_undef) {
						fprintf(stderr,"%s_%c undefined, skipping!\n",ent->u.track.dict,*s);
					} else {
						flatten_ent(sdict, ent2, tick_ofs, args, recur+1);
					}
				}
				tick_ofs += ent->u.track.ticks_per;
				s++;
			}
		}	break;
		case tde_command:
			add_event(tick_ofs,
				  ent->u.command.noteup,
			          arg_replace(ent->u.command.command, args));
			break;
	}
	if (ent->kind == tde_track) {
	} else if (ent->kind == tde_command) {
	}
}

void
tracker_show_status(void)
{
	unsigned count = 0;
	struct noteup_rec *n;

	printf("  timesteps per tick: %u\n", curr_per_tick);
	printf("  current tick      : %u\n", curr_tick);
	printf("  next tick in      : %u\n", curr_countdown);
	if (curr_event) {
		printf("  next event at tick: %u\n",curr_event->tick_ofs);
		printf("  next event        : %s\n",curr_event->command);
	}
	for (n = noteups; n; n = n->next) count++;
	printf("  pending note ups  : %u\n", count);
}

/* return 1 if the tracker is currently playing */
int
tracker_playing(void)
{
	return ((curr_event || noteups) ? 1 : 0);
}

void
tracker_stop(void)
{
	free_dicts();
	curr_tick = 0;
	curr_countdown = 0;
}
