/* Copyright 2010 Greg Alexander.  See the file COPYING for terms */
/* $Id: tree.c,v 1.46 2010-02-08 03:17:28 greg Exp $ */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "lang.h"
#include "misc.h"
#include "snd.h"

/* the maximum number of items to appear on any given comma-separated
 * list, such as formals or term_seq: */
#define MAX_LIST 100

static int current_func = 1;	/* just a unique identifier */

/* syntax:

   statement: sym fn_def ; | sym fn_call | expr ;
   fn_def:    [ formals ] = expr
   formals:   sym ,*
   expr:      factor + expr | factor - expr | factor
   factor:    value * factor | value / factor | value
   value:     sym | number | - number | ( expr ) | sym fn_call | term_seq | str | { expr , expr }
   sym:       t* | call to zero-arg function
   term_seq:  [ term_seq1 ,* ] | [ term_seq1 ,* ] :
   term_seq1: expr : expr | expr :: expr     :: is zero-crossing transition
   fn_call:   ( arg_list )
   arg_list:  expr ,*
*/

struct tree_rec *
new_tree(tree_op kind)
{
	struct tree_rec *tr;
	tr = (struct tree_rec *)safe_malloc(sizeof *tr);
	memset(tr,0,sizeof *tr);
	tr->kind = kind;
	return tr;
}


void
free_tree(struct tree_rec *tr)
{
	if (!tr)
		return;
	switch (tr->kind) {
		case op_error:
			break;
		case op_fn_def:
			if (tr->u.fn_def.name->defn == tr) {
				tr->u.fn_def.name->defn = NULL;
				if (tr->u.fn_def.name->compiled) {

					/* make sure no one is actively
					 * depending on this builtin */
					snd_shutup();

					compiled_free(tr->u.fn_def.name->compiled);
					tr->u.fn_def.name->compiled = NULL;
				}
			}
			free(tr->u.fn_def.formals);
			free_tree(tr->u.fn_def.expr);
			break;
		case op_fn_call: {
			int i;
			for (i = 0; i < tr->u.fn_call.num_args; i++) {
				free_tree(tr->u.fn_call.args[i]);
			}
			free(tr->u.fn_call.args);
		}	break;
		case op_add:
		case op_subtract:
		case op_multiply:
		case op_divide:
			free_tree(tr->u.operands[0]);
			free_tree(tr->u.operands[1]);
			break;
		case op_termseq: {
			int i;
			for (i = 0; i < tr->u.termseq.num_intervals; i++) {
				free_tree(tr->u.termseq.durations[i]);
				free_tree(tr->u.termseq.exprs[i]);
			}
			free(tr->u.termseq.durations);
			free(tr->u.termseq.exprs);
			free(tr->u.termseq.zero_crossing);
		}	break;
		case op_number:
			break;
		case op_argref:
			break;
		case op_string:
			break;
		case op_timest:
			free_tree(tr->u.timest.operand);
			break;
		case op_stereo:
			free_tree(tr->u.stereo.pan);
			free_tree(tr->u.stereo.value);
			break;
		case op_one_tree:
			free_tree(tr->u.one_tree.expr);
			free(tr->u.one_tree.tmps);
			break;
		case op_raised: {
			int i;
			free_tree(tr->u.raised.expr);
			for (i = 0; i < tr->u.raised.num_formals; i++) {
				free_tree(tr->u.raised.formals[i]);
			}
			free(tr->u.raised.formals);
			for (i = 0; i < tr->u.raised.num_commons; i++) {
				free_tree(tr->u.raised.commons[i]);
			}
			free(tr->u.raised.commons);
		}	break;
			
	}
	memset(tr,0,sizeof *tr);
	free(tr);
}

struct tree_rec *
dup_tree(struct tree_rec *tr)
{
	struct tree_rec *ret;
	if (!tr)
		return NULL;
	ret = new_tree(tr->kind);
	switch (tr->kind) {
		case op_error: break;
		case op_fn_def:
			assert_str(0,"fn_def in dup_tree() is considered weird");
			break;
		case op_fn_call: {
			int i,n;
			ret->u.fn_call.func = tr->u.fn_call.func;
			n = tr->u.fn_call.num_args;
			ret->u.fn_call.num_args = n;
			ret->u.fn_call.args = safe_malloc(n * sizeof *(ret->u.fn_call.args));
			for (i = 0; i < n; i++) {
				ret->u.fn_call.args[i] = dup_tree(tr->u.fn_call.args[i]);
			}
		}	break;
		case op_add:
		case op_subtract:
		case op_multiply:
		case op_divide:
			ret->u.operands[0] = dup_tree(tr->u.operands[0]);
			ret->u.operands[1] = dup_tree(tr->u.operands[1]);
			break;
		case op_termseq: {
			int i,n;
			n = tr->u.termseq.num_intervals;
			ret->u.termseq.num_intervals = n;
			ret->u.termseq.durations = safe_malloc(n * sizeof *(ret->u.termseq.durations));
			ret->u.termseq.exprs = safe_malloc(n * sizeof *(ret->u.termseq.exprs));
			for (i = 0; i < n; i++) {
				ret->u.termseq.durations[i] = dup_tree(tr->u.termseq.durations[i]);
				ret->u.termseq.exprs[i] = dup_tree(tr->u.termseq.exprs[i]);
			}
			ret->u.termseq.zero_crossing = safe_malloc(n * sizeof *(ret->u.termseq.zero_crossing));
			memcpy(ret->u.termseq.zero_crossing,tr->u.termseq.zero_crossing,n * sizeof *(ret->u.termseq.zero_crossing));
		}	break;
		case op_number:
			ret->u.number = tr->u.number;
			break;
		case op_string:
			ret->u.string = tr->u.string;
			break;
		case op_argref:
			ret->u.argref = tr->u.argref;
			break;
		case op_timest:
			ret->u.timest.operand = dup_tree(tr->u.timest.operand);
			ret->u.timest.kind = tr->u.timest.kind;
			break;
		case op_stereo:
			ret->u.stereo.pan = dup_tree(tr->u.stereo.pan);
			ret->u.stereo.value = dup_tree(tr->u.stereo.value);
			break;
		case op_one_tree:
			assert_str(0,"it is weird to copy one_tree");
			break;
		case op_raised:
			assert_str(0,"it is weird to copy op_raised");
			break;
	}
	return ret;
}


static struct tree_rec *scan_fn_def(struct symbol *sym);
static struct tree_rec *scan_fn_call(struct symbol *sym);
static struct tree_rec *scan_expr();
static struct tree_rec *scan_factor();
static struct tree_rec *scan_value();
static struct tree_rec *scan_term_seq(void);


/* statement: sym fn_def ; | sym fn_call | expr ;
 */
struct tree_rec *
scan_statement(void)
{
	struct symbol *sym;
	struct tree_rec *tr;

	next_token();
	if (curr_token == tok_eof) {
		return NULL;
	}
	if (curr_token == tok_symbol) {
		sym = curr_token_sym;
		next_token();
		if (curr_token == tok_lbracket) {
			tr = scan_fn_def(sym);
		} else if (curr_token == tok_lparen) {
			tr = scan_fn_call(sym);
		} else {
			do_error("expected fn call or definition (got %s)",dump_token());
			return new_tree(op_error);
		}
		if (curr_token != tok_semicolon) {
			do_warning("expected semicolon (got %s)",dump_token());
		}
	} else {
		tr = scan_expr();
	}
	if (num_errors != 0) {
		free_tree(tr);
		return new_tree(op_error);
	}
	return tr;
}


/* fn_def:    [ formals ] = expr
 * formals:   sym ,*
 */
static struct tree_rec *
scan_fn_def(struct symbol *sym)
{
	struct symbol *formals[MAX_LIST];
	int nformals = 0;
	struct tree_rec *tr;

	current_func++;

	assert_str(curr_token == tok_lbracket, "expected lbracket");

	tr = new_tree(op_error);

	/* scan formals */
	next_token();
	if (curr_token != tok_rbracket) {
		while (1) {
			if (curr_token != tok_symbol) {
				do_error("expected a symbol (got %s)",dump_token());
				return tr;
			}
			curr_token_sym->formal_index = nformals;
			curr_token_sym->formal_func = current_func;
			formals[nformals++] = curr_token_sym;
			if (nformals >= MAX_LIST) {
				do_error("too many formals");
				return tr;
			}
			next_token();
			if (curr_token != tok_comma) break;
			next_token();
		}
	}

	if (curr_token != tok_rbracket) {
		do_error("expected a closing bracket (got %s)",dump_token());
		return tr;
	}

	/* scan = */
	next_token();
	if (curr_token != tok_equals) {
		do_error("expecting assignment operator (got %s)",dump_token());
		return tr;
	}

	tr->kind = op_fn_def;
	tr->u.fn_def.name = sym;
	tr->u.fn_def.formals = safe_malloc(nformals * sizeof(struct symbol *));
	memcpy(tr->u.fn_def.formals,formals,nformals * sizeof(struct symbol *));
	tr->u.fn_def.num_formals = nformals;

	next_token();

	/* scan expr */
	tr->u.fn_def.expr = scan_expr();

	if (num_errors == 0) {
		if (!sym->defn && sym->builtin) {
			do_error("trying to override builtin %s",sym->name);
			free_tree(tr);
			return new_tree(op_error);
		}
		if (sym->defn) {
			if (nag_warnings) {
				do_warning("redefinition of %s",sym->name);
			}

			if (sym->compiled) {
				compiled_free(sym->compiled);
				sym->compiled = NULL;
			}

			/* make sure no one is actively
			 * depending on this builtin */
			snd_shutup();
			/* and same for compiled code: */
			reset_compileds();

			free_tree(sym->defn);
			sym->defn = NULL;
		}
		sym->defn = tr;
	}

	return tr;
}

/* fn_call:   ( arg_list )
 * arg_list:  expr ,*
 */
static struct tree_rec *
scan_fn_call(struct symbol *sym)
{
	struct tree_rec *tr = new_tree(op_error);
	struct tree_rec *args[MAX_LIST];
	int nargs = 0;

	/* scan arg_list */
	if (curr_token != tok_lparen) {
		do_error("expected left paren (got %s)",dump_token());
		return tr;
	}

	next_token();
	if (curr_token != tok_rparen) {
		while (1) {
			args[nargs++] = scan_expr();
			if (nargs >= MAX_LIST) {
				do_error("too many args");
				return tr;
			}
			if (curr_token != tok_comma) break;
			next_token();
		}
	}

	if (curr_token != tok_rparen) {
		do_error("expected right paren (got %s)",dump_token());
		return tr;
	}

	next_token();

	tr->kind = op_fn_call;
	tr->u.fn_call.func = sym;
	tr->u.fn_call.args = safe_malloc(nargs * sizeof(struct tree_rec *));
	memcpy(tr->u.fn_call.args,args,nargs * sizeof(struct tree_rec *));
	tr->u.fn_call.num_args = nargs;

	return tr;
}

/* expr:      factor + expr | factor - expr | factor
 */
static struct tree_rec *
scan_expr()
{
	struct tree_rec *tr, *tr_op;
	tr_op = scan_factor();
	if (curr_token == tok_plus || curr_token == tok_minus) {
		tr = new_tree((curr_token == tok_plus) ? op_add : op_subtract);
		tr->u.operands[0] = tr_op;
		next_token();
		tr->u.operands[1] = scan_expr();
	} else {
		tr = tr_op;
	}
	return tr;
}

/* factor:    value * factor | value / factor | value
 */
static struct tree_rec *
scan_factor()
{
	struct tree_rec *tr;
	struct tree_rec *tr_op0,*tr_op1;

	tr_op0 = scan_value();
	if (curr_token == tok_star || curr_token == tok_slash) {
		tree_op kind = (curr_token == tok_star) ? op_multiply
		                                        : op_divide;
		next_token();
		tr_op1 = scan_factor();
		if ((tr_op0->kind == op_timest || tr_op1->kind == op_timest) &&
		    (tr_op0->kind != tr_op1->kind)) {
			struct tree_rec *trx;
			tr = new_tree(kind);
			tr->u.operands[0] = (tr_op0->kind == op_timest) ? tr_op0->u.timest.operand : tr_op0;
			tr->u.operands[1] = (tr_op1->kind == op_timest) ? tr_op1->u.timest.operand : tr_op1;
			trx = (tr_op0->kind == op_timest) ? tr_op0 : tr_op1;
			trx->u.timest.operand = tr;
			tr = trx;
		} else {
			tr = new_tree(kind);
			tr->u.operands[0] = tr_op0;
			tr->u.operands[1] = tr_op1;
		}
	} else {
		tr = tr_op0;
	}
	return tr;
}

/* value:     sym | number | - number | ( expr ) | sym fn_call | term_seq | str | { expr , expr }
 */
static struct tree_rec *
scan_value()
{
	struct tree_rec *tr;

	if (curr_token == tok_symbol) {
	/* scan sym and fn_call */
		struct symbol *sym = curr_token_sym;
		next_token();
		if (curr_token == tok_lparen) {
			return scan_fn_call(sym);
		} else {
			if (sym->formal_func == current_func) {
				/* argument reference */
				tr = new_tree(op_argref);
				tr->u.argref = sym->formal_index;
			} else if (is_t_sym(sym) != t_none) {
				/* special "t" symbol reference */
				tr = new_tree(op_timest);
				tr->u.timest.operand = new_tree(op_number);
				tr->u.timest.operand->u.number = 1.0;
				tr->u.timest.kind = is_t_sym(sym);
			} else {
				/* build up a zero-arg function call,
				 * because that's how we do "global symbols" */
				tr = new_tree(op_fn_call);
				tr->u.fn_call.func = sym;
				tr->u.fn_call.args = NULL;
				tr->u.fn_call.num_args = 0;
			}
			return tr;
		}
	} else if (curr_token == tok_number) {
	/* scan number */
		tr = new_tree(op_number);
		tr->u.number = curr_token_num;
		next_token();
		return tr;
	} else if (curr_token == tok_minus) {
	/* scan - number */
		next_token();
		if (curr_token != tok_number) {
			do_error("expected number (got %s)",dump_token());
			return new_tree(op_error);
		}
		tr = new_tree(op_number);
		tr->u.number = -curr_token_num;
		next_token();
		return tr;
	} else if (curr_token == tok_lparen) {
	/* scan ( expr ) */
		next_token();
		tr = scan_expr();
		if (curr_token != tok_rparen) {
			do_error("expected right paren (got %s)",dump_token());
		} else {
			next_token();
		}
		return tr;
	} else if (curr_token == tok_lbracket) {
	/* scan term_seq */
		return scan_term_seq();
	} else if (curr_token == tok_string) {
	/* scan str */
		struct tree_rec *tr = new_tree(op_string);
		tr->u.string = add_string_to_strtab(curr_token_str);
		next_token();
		return tr;
	} else if (curr_token == tok_lbrace) {
	/* scan { expr , expr } */
		struct tree_rec *tr = new_tree(op_error);
		next_token();
		tr->u.stereo.pan = scan_expr();
		if (curr_token != tok_comma) {
			do_error("expected comma (got %s)",dump_token());
			free_tree(tr->u.stereo.pan);
			return tr;
		}
		next_token();
		tr->u.stereo.value = scan_expr();
		if (curr_token != tok_rbrace) {
			do_error("expected right brace (got %s)",dump_token());
			free_tree(tr->u.stereo.pan);
			free_tree(tr->u.stereo.value);
			return tr;
		}
		next_token();
		tr->kind = op_stereo;
		return tr;
	} else {
		do_error("looking for value (got %s)",dump_token());
	}
	return new_tree(op_error);
}

/* term_seq:  [ term_seq1 ,* ] | [ term_seq1 ,* ] :
 * term_seq1: expr : expr | expr :: expr
 */
static struct tree_rec *
scan_term_seq(void)
{
	struct tree_rec *tr = new_tree(op_error);
	struct tree_rec *durations[MAX_LIST];
	struct tree_rec *exprs[MAX_LIST];
	char zero_crossing[MAX_LIST];
	int nintervals = 0;

	do {
		/* skip the comma or lbracket */
		next_token();

		durations[nintervals] = scan_expr();

		if ((curr_token != tok_colon) &&
		    (curr_token != tok_coloncolon)) {
			do_error("expected colon (got %s)",dump_token());
			return tr;
		}
		zero_crossing[nintervals] = (curr_token == tok_coloncolon);

		next_token();

		exprs[nintervals] = scan_expr();

		nintervals++;
	} while (curr_token == tok_comma);

	tr->kind = op_termseq;
	tr->u.termseq.durations = safe_malloc(nintervals * sizeof(struct tree_rec *));
	memcpy(tr->u.termseq.durations,durations,nintervals * sizeof(struct tree_rec *));
	tr->u.termseq.exprs = safe_malloc(nintervals * sizeof(struct tree_rec *));
	memcpy(tr->u.termseq.exprs,exprs,nintervals * sizeof(struct tree_rec *));
	tr->u.termseq.zero_crossing = safe_malloc(nintervals * sizeof *(tr->u.termseq.zero_crossing));
	memcpy(tr->u.termseq.zero_crossing,zero_crossing,nintervals * sizeof *(tr->u.termseq.zero_crossing));
	tr->u.termseq.num_intervals = nintervals;
	tr->u.termseq.repeats = 0;

	if (curr_token != tok_rbracket) {
		do_error("expecting right bracket (got %s)",dump_token());
		free_tree(tr);
		return new_tree(op_error);
	}
	next_token();

	if (curr_token == tok_colon) {
		tr->u.termseq.repeats = 1;
		next_token();
	}

	return tr;
}


void
dump_tree(struct tree_rec *tr)
{
	if (!tr) { printf("NULL"); return; }
	switch(tr->kind) {
		case op_fn_def: {
			int i;
			printf("(def %s(", tr->u.fn_def.name->name);
			for (i = 0; i < tr->u.fn_def.num_formals; i++) {
				printf("%s ",tr->u.fn_def.formals[i]->name);
			}
			printf(") ");
			dump_tree(tr->u.fn_def.expr);
			printf(")");
		}	break;
		case op_fn_call: {
			int i;
			printf("(%s",tr->u.fn_call.func->name);
			for (i = 0; i < tr->u.fn_call.num_args; i++) {
				printf(" ");
				dump_tree(tr->u.fn_call.args[i]);
			}
			printf(")");
		}	break;
		case op_add:
		case op_subtract:
		case op_multiply:
		case op_divide:
			printf("(");
			dump_tree(tr->u.operands[0]);
			printf(" %c ", ((tr->kind == op_add) ? '+' : ((tr->kind == op_subtract) ? '-' : ((tr->kind == op_multiply) ? '*' : '/'))));
			dump_tree(tr->u.operands[1]);
			printf(")");
			break;
		case op_timest:
			printf("(t_%d",tr->u.timest.kind);
			printf(" * ");
			dump_tree(tr->u.timest.operand);
			printf(")");
			break;
		case op_termseq: {
			int i;
			printf("[");
			for (i = 0; i < tr->u.termseq.num_intervals; i++) {
				dump_tree(tr->u.termseq.durations[i]);
				printf(": ");
				dump_tree(tr->u.termseq.exprs[i]);
				printf(", ");
			}
			printf("]");
		}	break;
		case op_number:
			printf("%.3f",tr->u.number);
			break;
		case op_argref:
			if (tr->u.argref & ARG_RAISED_FORMAL) {
				printf("rform%d",tr->u.argref&~ARG_RAISED_FORMAL);
			} else if (tr->u.argref & ARG_RAISED_COMMON) {
				printf("rcom%d",tr->u.argref&~ARG_RAISED_COMMON);
			} else {
				printf("arg%d",tr->u.argref);
			}
			break;
		case op_string:
			printf("\"%s\"",get_string_from_strtab(tr->u.string));
			break;
		case op_stereo:
			printf("{");
			dump_tree(tr->u.stereo.pan);
			printf(",");
			dump_tree(tr->u.stereo.value);
			printf("}");
			break;
		case op_one_tree: {
			int i;
			
			printf("tmps=[");
			for (i = 0; i < tr->u.one_tree.num_temps; i++) {
				printf("%f ",(value_t)tr->u.one_tree.tmps[i]);
			}
			printf("], itmps=[");
			for (i = 0; i < tr->u.one_tree.num_itemps; i++) {
				printf("%d ",(intval_t)tr->u.one_tree.itmps[i]);
			}
			printf("]: (");
			dump_tree(tr->u.one_tree.expr);
			printf(")");
		}	break;
		case op_error:
			printf("error");
			break;
		case op_raised: {
			int i;
			printf("(raised[");
			for (i = 0; i < tr->u.raised.num_formals; i++) {
				if (i != 0) { printf(", "); }
				printf("f%d: ", i);
				dump_tree(tr->u.raised.formals[i]);
			}
			printf("; ");
			for (i = 0; i < tr->u.raised.num_commons; i++) {
				if (i != 0) { printf(", "); }
				printf("c%d: ", i);
				dump_tree(tr->u.raised.commons[i]);
			}
			printf("]: ");
			dump_tree(tr->u.raised.expr);
			printf(")");
		}	break;
	}
}


/* returns 1 iff x and y represent identical trees */
/* NOTE!!! This assumes that all argrefs refer to the same group of formals!!
 * So only use it on flattened trees or if you are certain they're at the
 * same function scope. */
int
tree_equal(struct tree_rec *x, struct tree_rec *y)
{
	if (x == y) return 1;		/* uh, weird shortcuts */
	if (!x || !y) return 0;
	if (x->kind != y->kind) return 0;
	switch (x->kind) {
		case op_error:
			return 0;
		case op_fn_def:
			return 0;
		case op_fn_call: {
			int i;
			if (x->u.fn_call.func != y->u.fn_call.func) return 0;
			if (x->u.fn_call.num_args != y->u.fn_call.num_args) return 0;
			for (i = 0; i < x->u.fn_call.num_args; i++) {
				if (!tree_equal(x->u.fn_call.args[i], y->u.fn_call.args[i])) return 0;
			}
			return 1;
		}	break;
		case op_add:
		case op_subtract:
		case op_multiply:
		case op_divide:
			return (tree_equal(x->u.operands[0],
					   y->u.operands[0]) &&
			        tree_equal(x->u.operands[1],
					   y->u.operands[1]));
		case op_termseq: {
			int i;
			if (x->u.termseq.num_intervals != y->u.termseq.num_intervals) return 0;
			if (memcmp(x->u.termseq.zero_crossing, y->u.termseq.zero_crossing, x->u.termseq.num_intervals)) return 0;
			if (x->u.termseq.repeats != y->u.termseq.repeats) return 0;
			for (i = 0; i < x->u.termseq.num_intervals; i++) {
				if (!tree_equal(x->u.termseq.durations[0], y->u.termseq.durations[0])) return 0;
				if (!tree_equal(x->u.termseq.exprs[0], y->u.termseq.exprs[0])) return 0;
			}
			return 1;
		}	break;
		case op_number:
			return (x->u.number == y->u.number);
		case op_argref:
			return (x->u.argref == y->u.argref);
		case op_string:
			return (x->u.string == y->u.string);
		case op_timest:
			if (x->u.timest.kind != y->u.timest.kind) return 0;
			return tree_equal(x->u.timest.operand, y->u.timest.operand);
		case op_stereo:
			return tree_equal(x->u.stereo.pan, y->u.stereo.pan) &&
			       tree_equal(x->u.stereo.value, y->u.stereo.value);
		case op_one_tree:	/* uh, this tree is already running? */
		case op_raised:	/* craziness! */
			return 0;
	}
	return 0;
}
